"""Common code for command line interface tool for the Jac language."""

from __future__ import annotations

import argparse
import cmd
import inspect
import pprint
import re
from typing import Callable, Dict, Optional


class Command:
    """Represents a command in the command line interface."""

    func: Callable
    sig: inspect.Signature

    def __init__(self, func: Callable) -> None:
        """Initialize a Command instance."""
        self.func = func
        self.sig = inspect.signature(func)

    def call(self, *args: list, **kwargs: dict) -> str:
        """Call the associated function with the specified arguments and keyword arguments."""
        return self.func(*args, **kwargs)


def extract_param_descriptions(docstring: str) -> Dict[str, str]:
    """Extract parameter descriptions from a function's docstring.

    Args:
        docstring: The function's docstring

    Returns:
        A dictionary mapping parameter names to their descriptions
    """
    param_descriptions: dict[str, str] = {}

    # Check if the docstring has an Args section
    args_match = re.search(r"Args:(.*?)(?:\n\n|\Z)", docstring, re.DOTALL)
    if not args_match:
        return param_descriptions

    args_section = args_match.group(1)

    # Extract parameter descriptions from the Args section
    # Looking for patterns like "param_name: Description" or "param_name (type): Description"
    current_param = None
    current_desc = []

    for line in args_section.strip().split("\n"):
        line = line.strip()
        if not line:
            continue

        # Match parameter name at the beginning of the line
        param_match = re.match(r"\s*([a-zA-Z0-9_]+)(?:\s*\([^)]*\))?:\s*(.*)", line)
        if param_match:
            # Save the previous parameter if there was one
            if current_param and current_desc:
                param_descriptions[current_param] = " ".join(current_desc)
                current_desc = []

            # Start a new parameter
            current_param = param_match.group(1)
            if param_match.group(2):
                current_desc.append(param_match.group(2))
        elif current_param and line:
            # Continue the description for the current parameter
            current_desc.append(line)

    # Save the last parameter
    if current_param and current_desc:
        param_descriptions[current_param] = " ".join(current_desc)

    return param_descriptions


class CommandRegistry:
    """Registry for managing commands in the command line interface."""

    registry: dict[str, Command]
    sub_parsers: argparse._SubParsersAction
    parser: argparse.ArgumentParser
    args: argparse.Namespace

    def __init__(self) -> None:
        """Initialize a CommandRegistry instance."""
        self.registry = {}
        self.parser = argparse.ArgumentParser(
            prog="jac",
            description="Jac Programming Language CLI - A tool for working with Jac programs",
            epilog="For more information, visit: https://github.com/Jaseci-Labs/jaseci",
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        self.parser.add_argument(
            "-V",
            "--version",
            action="store_true",
            help="Show the Jac version and installation path",
        )
        self.sub_parsers = self.parser.add_subparsers(
            title="commands",
            dest="command",
            description="The following commands are available:",
            metavar="COMMAND",
        )
        self.args = argparse.Namespace()

    def register(self, func: Callable) -> Callable:
        """Register a command in the registry."""
        name = func.__name__
        cmd = Command(func)
        self.registry[name] = cmd
        # Extract the first paragraph from the docstring for brief description
        doc = func.__doc__ or ""
        brief_desc = doc.split("\n\n")[0].strip()

        # Use the full docstring for the detailed description
        cmd_parser: argparse.ArgumentParser = self.sub_parsers.add_parser(
            name,
            description=func.__doc__,
            help=brief_desc,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        first = True
        for param_name, param in cmd.sig.parameters.items():
            # Get parameter type name
            type_name = (
                param.annotation.__name__
                if hasattr(param.annotation, "__name__")
                else str(param.annotation)
            )

            # Extract parameter descriptions from the docstring
            docstring_param_descriptions = extract_param_descriptions(doc)

            # Use docstring description if available, otherwise fallback to predefined or just type
            if param_name in docstring_param_descriptions:
                arg_msg = f"{type_name} - {docstring_param_descriptions[param_name]}"
            else:
                arg_msg = f"{type_name}"
            # shorthand is first character by default,
            # If already taken, use the first 2 characters
            shorthand = param_name[:1]
            if f"-{shorthand}" in cmd_parser._option_string_actions:
                shorthand = param_name[:2]
            if param_name == "args":
                cmd_parser.add_argument("args", nargs=argparse.REMAINDER, help=arg_msg)
            elif param_name == "filepath":
                first = False
                cmd_parser.add_argument(
                    f"{param_name}",
                    type=(
                        eval(param.annotation)
                        if isinstance(param.annotation, str)
                        else param.annotation
                    ),
                    help=arg_msg,
                    nargs="?",
                )
            elif param.default is param.empty:
                if first:
                    first = False
                    cmd_parser.add_argument(
                        f"{param_name}",
                        type=(
                            eval(param.annotation)
                            if isinstance(param.annotation, str)
                            else param.annotation
                        ),
                        help=arg_msg,
                    )
                else:
                    cmd_parser.add_argument(
                        f"-{shorthand}",
                        f"--{param_name}",
                        required=True,
                        type=(
                            eval(param.annotation)
                            if isinstance(param.annotation, str)
                            else param.annotation
                        ),
                        help=arg_msg,
                    )
            elif first:
                arg_msg += f", default: {param.default}"
                first = False
                cmd_parser.add_argument(
                    f"{param_name}",
                    default=param.default,
                    type=eval(param.annotation),
                    help=arg_msg,
                )
            else:
                arg_msg += f", default: {param.default}"
                if param.annotation == bool:
                    cmd_parser.add_argument(
                        f"-{shorthand}",
                        f"--{param_name}",
                        default=param.default,
                        action="store_true",
                        help=arg_msg,
                    )
                    cmd_parser.add_argument(
                        f"-n{shorthand}",
                        f"--no-{param_name}",
                        dest=param_name,
                        action="store_false",
                        help=f"Compliment of {arg_msg}",
                    )
                else:
                    cmd_parser.add_argument(
                        f"-{shorthand}",
                        f"--{param_name}",
                        default=param.default,
                        help=arg_msg,
                        type=(
                            eval(param.annotation)
                            if isinstance(param.annotation, str)
                            else param.annotation
                        ),
                    )
        return func

    def get(self, name: str) -> Optional[Command]:
        """Get the Command instance for a given command name."""
        return self.registry.get(name)

    def get_all_commands(self) -> dict:
        """Get all registered commands along with their details."""
        all_commands = {}
        for name, comd in self.registry.items():
            doc = comd.func.__doc__ or "No help available."
            args = comd.sig.parameters
            all_commands[name] = (doc, args)
        return all_commands


cmd_registry = CommandRegistry()


class CommandShell(cmd.Cmd):
    """Command shell for the command line interface."""

    intro = "Welcome to the Jac CLI!"
    prompt = "jac> "
    cmd_reg: CommandRegistry

    def __init__(self, cmd_reg: CommandRegistry) -> None:
        """Initialize a CommandShell instance."""
        self.cmd_reg = cmd_reg
        super().__init__()

    def do_exit(self, arg: list) -> bool:
        """Exit the command shell."""
        return True

    def default(self, line: str) -> None:
        """Process the command line input."""
        args = vars(self.cmd_reg.parser.parse_args(line.split()))
        command = self.cmd_reg.get(args["command"])
        if command:
            args.pop("command")
            ret = command.call(**args)
            if ret:
                ret_str = pprint.pformat(ret, indent=2)
                self.stdout.write(f"{ret_str}\n")

    def do_help(self, arg: str) -> None:
        """Jac CLI help implementation."""

        def get_info(name: str, doc: str, args: dict[str, inspect.Parameter]) -> None:
            """Format and display detailed command information."""
            # Format the command header
            self.stdout.write(f"\n{'=' * 80}\n")
            self.stdout.write(f"COMMAND: {name}\n")
            self.stdout.write(f"{'=' * 80}\n\n")

            # Format the command description
            doc_lines = doc.strip().split("\n")
            for line in doc_lines:
                self.stdout.write(f"{line}\n")

            # Format the command arguments
            if args:
                self.stdout.write("\nARGUMENTS:\n")
                for param_name, param in args.items():
                    # Get parameter type
                    type_name = (
                        param.annotation.__name__
                        if hasattr(param.annotation, "__name__")
                        else str(param.annotation)
                    )

                    # Format default value if present
                    default_str = ""
                    if param.default is not param.empty:
                        default_str = f" (default: {param.default})"

                    # Format required/optional status
                    req_str = (
                        " [required]"
                        if param.default is param.empty and param_name != "args"
                        else " [optional]"
                    )

                    self.stdout.write(
                        f"  {param_name}: {type_name}{default_str}{req_str}\n"
                    )
            else:
                self.stdout.write("\nNo arguments\n")

            # Format usage examples
            command_parser = self.cmd_reg.sub_parsers.choices[name]
            self.stdout.write(f"\nUSAGE:\n  {command_parser.format_usage()[7:]}\n")

            # Extract and format examples from docstring if present
            if "Examples:" in doc:
                examples_section = doc.split("Examples:")[1].strip()
                example_lines = examples_section.split("\n")
                self.stdout.write("\nEXAMPLES:\n")
                for example in example_lines:
                    if example.strip():
                        self.stdout.write(f"  {example.strip()}\n")

            self.stdout.write("\n")

        if arg == "all":
            # Display all commands with their details
            command_details = self.cmd_reg.get_all_commands()
            for name, (doc, args) in sorted(command_details.items()):
                get_info(name, doc, args)

        elif arg:
            # Display help for a specific command
            command = self.cmd_reg.get(arg)
            if command:
                doc = command.func.__doc__ or "No help available."
                args = command.sig.parameters
                get_info(arg, doc, args)
            else:
                self.stdout.write(f"\nUnknown command: '{arg}'\n")
                self.stdout.write("Type 'help' to see available commands.\n")
        else:
            # Display general help information
            self.stdout.write("\n")
            self.stdout.write("JAC PROGRAMMING LANGUAGE COMMAND LINE INTERFACE\n")
            self.stdout.write("=" * 50 + "\n\n")

            self.stdout.write("AVAILABLE COMMANDS:\n")

            # Get all command names and sort them alphabetically
            command_names = sorted(self.cmd_reg.registry.keys())

            # Get brief descriptions for each command
            command_descriptions: dict[str, str] = {}
            for name in command_names:
                command = self.cmd_reg.get(name)
                if command and command.func.__doc__:
                    # Extract the first line of the docstring as a brief description
                    brief = command.func.__doc__.split("\n")[0].strip()
                    command_descriptions[name] = brief
                else:
                    command_descriptions[name] = "No description available"

            # Display commands with their brief descriptions
            for name in command_names:
                self.stdout.write(
                    f"  {name.ljust(15)} - {command_descriptions[name]}\n"
                )

            self.stdout.write("\nFor detailed information on a specific command:\n")
            self.stdout.write("  help <command>\n")

            self.stdout.write("\nTo see detailed information for all commands:\n")
            self.stdout.write("  help all\n")

            self.stdout.write("\nTo exit the Jac CLI:\n")
            self.stdout.write("  exit\n\n")



from _typeshed import Incomplete

from authlib.jose.rfc7517 import AsymmetricKey

PUBLIC_KEYS_MAP: Incomplete
PRIVATE_KEYS_MAP: Incomplete

class OKPKey(AsymmetricKey):
    kty: str
    REQUIRED_JSON_FIELDS: Incomplete
    PUBLIC_KEY_FIELDS = REQUIRED_JSON_FIELDS
    PRIVATE_KEY_FIELDS: Incomplete
    PUBLIC_KEY_CLS: Incomplete
    PRIVATE_KEY_CLS: Incomplete
    SSH_PUBLIC_PREFIX: bytes
    def exchange_shared_key(self, pubkey): ...
    @staticmethod
    def get_key_curve(key): ...
    def load_private_key(self): ...
    def load_public_key(self): ...
    def dumps_private_key(self): ...
    def dumps_public_key(self, public_key=None): ...
    @classmethod
    def generate_key(cls, crv: str = "Ed25519", options=None, is_private: bool = False) -> OKPKey: ...



from typing import Any

from braintree.graphql.unions.customer_recommendations import CustomerRecommendations

class CustomerRecommendationsPayload:
    is_in_paypal_network: bool | None
    recommendations: CustomerRecommendations | None
    def __init__(
        self,
        is_in_paypal_network: bool | None = None,
        recommendations: CustomerRecommendations | None = None,
        response: dict[str, Any] | None = None,
    ): ...



"""Test pass module."""

from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCase
import unittest


class TestCFGBuildPass(TestCase):
    """Test FuseTypeInfoPass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_cfg_branches_and_loops(self) -> None:
        """Test basic blocks."""
        file_name = self.fixture_abs_path("cfg_gen.jac")

        from jaclang.compiler.passes.main.cfg_build_pass import CoalesceBBPass

        with open(file_name, "r") as f:
            file_source = f.read()

        ir = (prog := JacProgram()).compile(use_str=file_source, file_path=file_name)

        cfg_pass = CoalesceBBPass(
            ir_in=ir,
            prog=prog,
        )

        dot = cfg_pass.printgraph_cfg()

        expected_dot = (
            "digraph G {\n"
            '  0 [label="BB0\\nx = 5 ;\\ny = 3 ;\\nif x > 3", shape=box];\n'
            '  1 [label="BB1\\nz = x + y ;", shape=box];\n'  # Note double backslash
            '  2 [label="BB2\\nelif x == 0", shape=box];\n'
            '  3 [label="BB3\\nz = y ;", shape=box];\n'
            '  4 [label="BB4\\nz = x - y ;", shape=box];\n'
            '  5 [label="BB5\\nfor i in range ( 0 , 10 )", shape=box];\n'
            '  6 [label="BB6\\nz = z + 1 ;", shape=box];\n'
            '  7 [label="BB7\\nwhile z < 10\\nz = z + 2 ;", shape=box];\n'
            '  8 [label="BB8\\nz = z + 2 ;", shape=box];\n'
            "  0 -> 1;\n"
            "  0 -> 2;\n"
            "  1 -> 5;\n"
            "  2 -> 3;\n"
            "  2 -> 4;\n"
            "  3 -> 5;\n"
            "  4 -> 5;\n"
            "  5 -> 6;\n"
            "  5 -> 7;\n"
            "  6 -> 5;\n"
            "  7 -> 7;\n"
            "  8 -> 7;\n"
            "}\n"
        )

        self.assertEqual(dot, expected_dot)

    def test_cfg_abilities_and_objects(self) -> None:
        """Test basic blocks."""
        file_name = self.fixture_abs_path("cfg_ability_test.jac")

        from jaclang.compiler.passes.main.cfg_build_pass import CoalesceBBPass

        with open(file_name, "r") as f:
            file_source = f.read()

        ir = (prog := JacProgram()).compile(use_str=file_source, file_path=file_name)

        cfg_pass = CoalesceBBPass(
            ir_in=ir,
            prog=prog,
        )

        dot = cfg_pass.printgraph_cfg()

        expected_dot = (
            "digraph G {\n"
            '  0 [label="BB0\\nobj math_mod", shape=box];\n'
            '  1 [label="BB1\\ncan divide( x : int , y : int )\\nif y == 0", shape=box];\n'
            '  2 [label="BB2\\nreturn 0 ;", shape=box];\n'
            '  3 [label="BB3\\nreturn x / y ;", shape=box];\n'
            '  4 [label="BB4\\ncan multiply( x : int , y : int )\\nreturn x * y ;", shape=box];\n'
            '  5 [label="BB5\\nx = 5 ;\\ny = 0 ;\\nmath = math_mod ( ) ;\\nz = math . divide ( x , y ) '
            ';\\nprint ( z ) ;", shape=box];\n'
            "  0 -> 1;\n"
            "  0 -> 4;\n"
            "  1 -> 2;\n"
            "  1 -> 3;\n"
            "}\n"
        )

        self.assertEqual(dot, expected_dot)



"""Pass to fuse comments into the AST."""

import jaclang.compiler.unitree as uni
from jaclang.compiler.passes import UniPass


class FuseCommentsPass(UniPass):
    """Fuses comment tokens into the AST at appropriate positions."""

    def before_pass(self) -> None:
        self.all_tokens: list[uni.Token] = []
        self.comments: list[uni.CommentToken] = []
        if isinstance(self.ir_out, uni.Module):
            self.comments = self.ir_out.source.comments
        return super().before_pass()

    def exit_node(self, node: uni.UniNode) -> None:
        if isinstance(node, uni.Token):
            self.all_tokens.append(node)

    def after_pass(self) -> None:
        # Early exit if no comments
        if not self.comments:
            return
        # Merge comments and code tokens in correct order
        merged_tokens = self._merge_tokens()
        # Insert comments into the AST structure
        self._insert_comments_in_ast(merged_tokens)

    def _merge_tokens(self) -> list[uni.Token]:
        """Merge comments and code tokens in correct order based on position."""
        merged: list[uni.Token] = []
        code_tokens = iter(self.all_tokens)
        comments = iter(self.comments)
        try:
            next_code = next(code_tokens)
        except StopIteration:
            next_code = None
        try:
            next_comment = next(comments)
        except StopIteration:
            next_comment = None

        # Handle possible leading comments
        if next_comment and (not next_code or _is_before(next_comment, next_code)):
            self.ir_out.src_terminals.insert(0, next_comment)

        # Merge streams in order
        while next_comment or next_code:
            if next_comment and (not next_code or _is_before(next_comment, next_code)):
                # Add comment token
                if merged and (last_token := merged[-1]):
                    self.ir_out.src_terminals.insert(
                        self.ir_out.src_terminals.index(last_token) + 1,
                        next_comment,
                    )
                merged.append(next_comment)
                try:
                    next_comment = next(comments)
                except StopIteration:
                    next_comment = None
            elif next_code:  # Check that next_code is not None
                # Add code token
                merged.append(next_code)
                try:
                    next_code = next(code_tokens)
                except StopIteration:
                    next_code = None

        return merged

    def _insert_comments_in_ast(self, merged_tokens: list[uni.Token]) -> None:
        """Insert comment tokens into the appropriate places in the AST."""
        i = 0
        while i < len(merged_tokens):
            token = merged_tokens[i]
            if not isinstance(token, uni.CommentToken):
                i += 1
                continue

            # Start collecting consecutive comments
            comment_batch = [token]
            next_idx = i + 1

            # Gather consecutive comments
            while next_idx < len(merged_tokens) and isinstance(
                (next_cmt := merged_tokens[next_idx]), uni.CommentToken
            ):
                comment_batch.append(next_cmt)
                next_idx += 1

            if next_idx >= len(merged_tokens):
                # Last tokens are comments - add batch to end of tree
                self.ir_out.add_kids_right(comment_batch)
            else:
                # Insert before the next non-comment token in its parent's children
                next_token = merged_tokens[next_idx]
                if next_token.parent is None:
                    raise self.ice(
                        f"Token {next_token.pp()} without parent in AST while"
                        f" inserting comments batch"
                    )
                parent_node = next_token.parent
                insert_index = parent_node.kid.index(next_token)
                parent_node.insert_kids_at_pos(comment_batch, insert_index)

            # Skip past all the comments we just processed
            i = next_idx


def _is_before(comment: uni.CommentToken, code: uni.Token) -> bool:
    """Determine if comment should come before the code token."""
    if comment.loc.first_line < code.loc.first_line:
        return True
    elif comment.loc.first_line == code.loc.first_line:
        comment.is_inline = True
        return comment.loc.col_start < code.loc.col_start
    return False



"""Test for Jac language server[VSCE] features"""

import os
import pytest

from lsprotocol.types import (
    DidOpenTextDocumentParams,
    TextDocumentItem,
    DidSaveTextDocumentParams,
    DidChangeTextDocumentParams,
    DocumentFormattingParams,
    TextEdit,
    VersionedTextDocumentIdentifier,
    TextDocumentIdentifier,
)
from jaclang.langserve.tests.server_test.utils import (
    create_temp_jac_file,
    get_code,
    get_jac_file_path,
    get_simple_code,
    create_ls_with_workspace,  # new helper
)
from jaclang.vendor.pygls.uris import from_fs_path
from jaclang.vendor.pygls.workspace import Workspace
from jaclang import JacMachineInterface as _
from jaclang.langserve.engine import JacLangServer
from jaclang.langserve.server import did_open, did_save, did_change, formatting

JAC_FILE = get_jac_file_path()


class TestLangServe:
    """Test class for Jac language server features."""

    @pytest.mark.asyncio
    async def test_open_valid_file_no_diagnostics(self):
        """Test opening a Jac file with a syntax error."""
        ls = JacLangServer()

        code = get_code("")
        temp_file_path = create_temp_jac_file(code)
        uri = from_fs_path(temp_file_path)
        ls.lsp._workspace = Workspace(os.path.dirname(temp_file_path), ls)

        params = DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=1,
                text=code,
            )
        )
        await did_open(ls, params)

        diagnostics = ls.diagnostics.get(uri, [])
        ls.shutdown()
        assert len(diagnostics) == 0

        os.remove(temp_file_path)

    @pytest.mark.asyncio
    async def test_open_with_syntax_error(self):
        """Test opening a Jac file with a syntax error."""
        ls = JacLangServer()

        code = get_code("error")
        temp_file_path = create_temp_jac_file(code)
        uri = from_fs_path(temp_file_path)
        ls.lsp._workspace = Workspace(os.path.dirname(temp_file_path), ls)

        params = DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=1,
                text=code,
            )
        )
        await did_open(ls, params)

        diagnostics = ls.diagnostics.get(uri, [])
        ls.shutdown()
        assert len(diagnostics) == 1
        assert diagnostics[0].message == "Syntax Error"
        assert str(diagnostics[0].range) == "66:0-66:1"

        os.remove(temp_file_path)

    @pytest.mark.asyncio
    async def test_did_open_and_simple_syntax_error(self):
        """Test diagnostics for a Jac file with a syntax error."""
        ls = JacLangServer()

        code = get_code("")
        temp_file_path = create_temp_jac_file(code)
        uri = from_fs_path(temp_file_path)
        ls.lsp._workspace = Workspace(os.path.dirname(temp_file_path), ls)

        params = DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=1,
                text=code,
            )
        )
        await did_open(ls, params)
        diagnostics = ls.diagnostics.get(uri, [])
        assert isinstance(diagnostics, list)
        assert len(diagnostics) == 0

        broken_code = get_code("error")
        with open(temp_file_path, "w") as f:
            f.write(broken_code)
        params = DidOpenTextDocumentParams(
            text_document=TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=1,
                text=broken_code,
            )
        )
        await did_open(ls, params)
        diagnostics = ls.diagnostics.get(uri, [])
        assert isinstance(diagnostics, list)
        assert len(diagnostics) == 1

        sem_tokens = ls.get_semantic_tokens(uri)
        print(sem_tokens)
        ls.shutdown()
        assert hasattr(sem_tokens, "data")
        assert isinstance(sem_tokens.data, list)
        assert (
            len(sem_tokens.data) == 0
        )  # TODO: we should retain the sem tokens, will be fixed in next PR

        os.remove(temp_file_path)

    @pytest.mark.asyncio
    async def test_did_save(self):
        """Test saving a Jac file triggers diagnostics."""
        code = get_code("")
        temp_file_path = create_temp_jac_file(code)
        uri, ls = create_ls_with_workspace(temp_file_path)

        await did_open(
            ls,
            DidOpenTextDocumentParams(
                text_document=TextDocumentItem(
                    uri=uri,
                    language_id="jac",
                    version=1,
                    text=code,
                )
            ),
        )

        params = DidSaveTextDocumentParams(
            text_document=TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=2,
                text=code,
            )
        )

        await did_save(ls, params)
        diagnostics = ls.diagnostics.get(uri, [])
        assert isinstance(diagnostics, list)
        assert len(diagnostics) == 0

        # Now simulate a syntax error by updating the workspace and saving
        broken_code = get_code("error")
        ls.workspace.put_text_document(
            TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=3,
                text=broken_code,
            )
        )
        params = DidSaveTextDocumentParams(
            text_document=TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=3,
                text=broken_code,
            )
        )
        await did_save(ls, params)
        sem_tokens = ls.get_semantic_tokens(uri)
        # semantic tokens should still be present even if there is a syntax error
        assert len(sem_tokens.data) == 300
        diagnostics = ls.diagnostics.get(uri, [])
        assert isinstance(diagnostics, list)
        assert len(diagnostics) == 1
        assert diagnostics[0].message == "Syntax Error"

        ls.shutdown()
        os.remove(temp_file_path)

    @pytest.mark.asyncio
    async def test_did_change(self):
        """Test changing a Jac file triggers diagnostics."""
        code = get_code("")
        temp_file_path = create_temp_jac_file(code)
        uri, ls = create_ls_with_workspace(temp_file_path)

        await did_open(
            ls,
            DidOpenTextDocumentParams(
                text_document=TextDocumentItem(
                    uri=uri,
                    language_id="jac",
                    version=1,
                    text=code,
                )
            ),
        )

        # No error, should be no diagnostics
        params = DidChangeTextDocumentParams(
            text_document=VersionedTextDocumentIdentifier(uri=uri, version=2),
            content_changes=[{"text": "\n" + code}],
        )
        ls.workspace.put_text_document(
            TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=2,
                text="\n" + code,
            )
        )
        await did_change(ls, params)
        diagnostics = ls.diagnostics.get(uri, [])
        assert isinstance(diagnostics, list)
        assert len(diagnostics) == 0

        # Now add a syntax error and update workspace
        # This should trigger diagnostics with a syntax error
        error_code = "\nerror"
        params = DidChangeTextDocumentParams(
            text_document=VersionedTextDocumentIdentifier(uri=uri, version=3),
            content_changes=[{"text": error_code + code}],
        )
        ls.workspace.put_text_document(
            TextDocumentItem(
                uri=uri,
                language_id="jac",
                version=3,
                text=error_code + code,
            )
        )
        await did_change(ls, params)
        sem_tokens = ls.get_semantic_tokens(uri)
        # semantic tokens should still be present even if there is a syntax error
        assert len(sem_tokens.data) == 300
        diagnostics = ls.diagnostics.get(uri, [])
        assert isinstance(diagnostics, list)
        assert len(diagnostics) == 1
        assert diagnostics[0].message == "Syntax Error"

        ls.shutdown()
        os.remove(temp_file_path)

    def test_vsce_formatting(self):
        """Test formatting a Jac file returns edits."""
        code = get_code("")
        temp_file_path = create_temp_jac_file(code)
        uri, ls = create_ls_with_workspace(temp_file_path)
        params = DocumentFormattingParams(
            text_document=TextDocumentIdentifier(uri=uri),
            options={"tabSize": 4, "insertSpaces": True},
        )
        edits = formatting(ls, params)
        assert isinstance(edits, list)
        assert isinstance(edits[0], TextEdit)
        assert (
            len(edits[0].new_text) > 100
        )  # it is a random number to check if the text is changed
        print(edits[0].new_text)
        ls.shutdown()
        os.remove(temp_file_path)

    @pytest.mark.asyncio
    async def test_multifile_workspace(self):
        """Test opening multiple Jac files in a workspace."""
        code1 = get_simple_code("")
        code2 = get_simple_code("error")
        temp_file_path1 = create_temp_jac_file(code1)
        temp_file_path2 = create_temp_jac_file(code2)

        uri1, ls = create_ls_with_workspace(temp_file_path1)
        uri2 = from_fs_path(temp_file_path2)

        await did_open(
            ls,
            DidOpenTextDocumentParams(
                text_document=TextDocumentItem(
                    uri=uri1,
                    language_id="jac",
                    version=1,
                    text=code1,
                )
            ),
        )
        await did_open(
            ls,
            DidOpenTextDocumentParams(
                text_document=TextDocumentItem(
                    uri=uri2,
                    language_id="jac",
                    version=1,
                    text=code2,
                )
            ),
        )

        diagnostics1 = ls.diagnostics.get(uri1, [])
        diagnostics2 = ls.diagnostics.get(uri2, [])
        assert len(diagnostics1) == 0
        assert len(diagnostics2) == 1
        assert diagnostics2[0].message == "Syntax Error"

        before_sem_tokens_1 = ls.get_semantic_tokens(uri1)
        before_sem_tokens_2 = ls.get_semantic_tokens(uri2)
        assert len(before_sem_tokens_1.data) == 15
        assert len(before_sem_tokens_2.data) == 0

        changed_code = get_simple_code("glob x = 90;")
        ls.workspace.put_text_document(
            TextDocumentItem(
                uri=uri1,
                language_id="jac",
                version=2,
                text=changed_code,
            )
        )
        params = DidChangeTextDocumentParams(
            text_document=VersionedTextDocumentIdentifier(uri=uri1, version=2),
            content_changes=[{"text": changed_code}],
        )
        await did_change(ls, params)

        after_sem_tokens_1 = ls.get_semantic_tokens(uri1)
        after_sem_tokens_2 = ls.get_semantic_tokens(uri2)

        assert len(after_sem_tokens_1.data) == 20
        assert len(after_sem_tokens_2.data) == 0

        ls.shutdown()
        os.remove(temp_file_path1)
        os.remove(temp_file_path2)



"""Core constructs for Jac Language."""

from __future__ import annotations


from .archetype import (
    AccessLevel,
    Anchor,
    Archetype,
    DataSpatialFunction,
    EdgeAnchor,
    EdgeArchetype,
    GenericEdge,
    NodeAnchor,
    NodeArchetype,
    Root,
    WalkerAnchor,
    WalkerArchetype,
)
from .memory import Memory, ShelfStorage
from .test import JacTestCheck, JacTestResult, JacTextTestRunner

__all__ = [
    "AccessLevel",
    "Anchor",
    "NodeAnchor",
    "EdgeAnchor",
    "WalkerAnchor",
    "Archetype",
    "NodeArchetype",
    "EdgeArchetype",
    "WalkerArchetype",
    "GenericEdge",
    "Root",
    "DataSpatialFunction",
    "Memory",
    "ShelfStorage",
    "JacTestResult",
    "JacTextTestRunner",
    "JacTestCheck",
]



import datetime
import sys
from _typeshed import Incomplete

if sys.platform == "win32":
    handle: Incomplete
    tzparent: Incomplete
    parentsize: Incomplete
    localkey: Incomplete
    WEEKS: Incomplete
    def list_timezones(): ...

    class win32tz(datetime.tzinfo):
        data: Incomplete
        def __init__(self, name) -> None: ...
        def utcoffset(self, dt): ...
        def dst(self, dt): ...
        def tzname(self, dt): ...

    def pickNthWeekday(year, month, dayofweek, hour, minute, whichweek): ...

    class win32tz_data:
        display: Incomplete
        dstname: Incomplete
        stdname: Incomplete
        stdoffset: Incomplete
        dstoffset: Incomplete
        stdmonth: Incomplete
        stddayofweek: Incomplete
        stdweeknumber: Incomplete
        stdhour: Incomplete
        stdminute: Incomplete
        dstmonth: Incomplete
        dstdayofweek: Incomplete
        dstweeknumber: Incomplete
        dsthour: Incomplete
        dstminute: Incomplete
        def __init__(self, path) -> None: ...

    def valuesToDict(key): ...



from _typeshed import Incomplete
from collections.abc import Mapping
from typing import ClassVar

__all__ = ["Dialog"]

class Dialog:
    command: ClassVar[str | None]
    master: Incomplete | None
    options: Mapping[str, Incomplete]
    def __init__(self, master=None, **options) -> None: ...
    def show(self, **options): ...



from _typeshed import StrOrBytesPath
from collections.abc import Sequence
from tarfile import TarInfo

def rebuild(
    filename: StrOrBytesPath, tag=None, format: str = "gz", zonegroups: Sequence[str | TarInfo] = [], metadata=None
) -> None: ...



"""Collection of passes for Jac IR."""

from ..transform import Alert, Transform  # noqa: I100
from .annex_pass import JacAnnexPass  # noqa: I100
from .sym_tab_build_pass import SymTabBuildPass, UniPass  # noqa: I100
from .sym_tab_link_pass import SymTabLinkPass  # noqa: I100
from .def_use_pass import DefUsePass  # noqa: I100
from .sem_def_match_pass import SemDefMatchPass  # noqa: I100
from .import_pass import JacImportDepsPass, PyImportDepsPass  # noqa: I100
from .def_impl_match_pass import DeclImplMatchPass  # noqa: I100
from .pyast_load_pass import PyastBuildPass  # type: ignore # noqa: I100
from .pyast_gen_pass import PyastGenPass  # noqa: I100
from .pybc_gen_pass import PyBytecodeGenPass  # noqa: I100
from .cfg_build_pass import CFGBuildPass  # noqa: I100
from .pyjac_ast_link_pass import PyJacAstLinkPass  # noqa: I100
from .inheritance_pass import InheritancePass  # noqa: I100


__all__ = [
    "Alert",
    "Transform",
    "UniPass",
    "JacAnnexPass",
    "JacImportDepsPass",
    "PyImportDepsPass",
    "SymTabBuildPass",
    "SymTabLinkPass",
    "DeclImplMatchPass",
    "DefUsePass",
    "SemDefMatchPass",
    "PyastBuildPass",
    "PyastGenPass",
    "PyBytecodeGenPass",
    "CFGBuildPass",
    "PyJacAstLinkPass",
    "InheritancePass",
]



"""Core constructs for Jac Language."""

from __future__ import annotations

from dataclasses import dataclass, field
from pickle import dumps
from shelve import Shelf, open
from typing import Any, Callable, Generator, Generic, Iterable, TypeVar, cast
from uuid import UUID

from .archetype import Anchor, NodeAnchor, Root, TANCH

ID = TypeVar("ID")


@dataclass
class Memory(Generic[ID, TANCH]):
    """Generic Memory Handler."""

    __mem__: dict[ID, TANCH] = field(default_factory=dict)
    __gc__: set[TANCH] = field(default_factory=set)

    def close(self) -> None:
        """Close memory handler."""
        self.__mem__.clear()
        self.__gc__.clear()

    def is_cached(self, id: ID) -> bool:
        """Check if id if already cached."""
        return id in self.__mem__

    def query(
        self, filter: Callable[[TANCH], bool] | None = None
    ) -> Generator[TANCH, None, None]:
        """Find anchors from memory with filter."""
        return (
            anchor for anchor in self.__mem__.values() if not filter or filter(anchor)
        )

    def all_root(self) -> Generator[Root, None, None]:
        """Get all the roots."""
        for anchor in self.query(lambda anchor: isinstance(anchor.archetype, Root)):
            yield cast(Root, anchor.archetype)

    def find(
        self,
        ids: ID | Iterable[ID],
        filter: Callable[[TANCH], TANCH] | None = None,
    ) -> Generator[TANCH, None, None]:
        """Find anchors from memory by ids with filter."""
        if not isinstance(ids, Iterable):
            ids = [ids]

        return (
            anchor
            for id in ids
            if (anchor := self.__mem__.get(id)) and (not filter or filter(anchor))
        )

    def find_one(
        self,
        ids: ID | Iterable[ID],
        filter: Callable[[TANCH], TANCH] | None = None,
    ) -> TANCH | None:
        """Find one anchor from memory by ids with filter."""
        return next(self.find(ids, filter), None)

    def find_by_id(self, id: ID) -> TANCH | None:
        """Find one by id."""
        return self.__mem__.get(id)

    def set(self, id: ID, data: TANCH) -> None:
        """Save anchor to memory."""
        self.__mem__[id] = data

    def remove(self, ids: ID | Iterable[ID]) -> None:
        """Remove anchor/s from memory."""
        if not isinstance(ids, Iterable):
            ids = [ids]

        for id in ids:
            if anchor := self.__mem__.pop(id, None):
                self.__gc__.add(anchor)


@dataclass
class ShelfStorage(Memory[UUID, Anchor]):
    """Shelf Handler."""

    __shelf__: Shelf[Anchor] | None = None

    def __init__(self, session: str | None = None) -> None:
        """Initialize memory handler."""
        super().__init__()
        self.__shelf__ = open(session) if session else None  # noqa: SIM115

    def close(self) -> None:
        """Close memory handler."""
        if isinstance(self.__shelf__, Shelf):
            for anchor in self.__gc__:
                self.__shelf__.pop(str(anchor.id), None)
                self.__mem__.pop(anchor.id, None)

            keys = set(self.__mem__.keys())

            # current memory
            self.sync_mem_to_db(keys)

            # additional after memory sync
            self.sync_mem_to_db(set(self.__mem__.keys() - keys))

            self.__shelf__.close()
        super().close()

    def sync_mem_to_db(self, keys: Iterable[UUID]) -> None:
        """Manually sync memory to db."""
        from jaclang.runtimelib.machine import JacMachineInterface as Jac

        if isinstance(self.__shelf__, Shelf):
            for key in keys:
                if (
                    (d := self.__mem__.get(key))
                    and d.persistent
                    and d.hash != hash(dumps(d))
                ):
                    _id = str(d.id)
                    if p_d := self.__shelf__.get(_id):
                        if (
                            isinstance(p_d, NodeAnchor)
                            and isinstance(d, NodeAnchor)
                            and p_d.edges != d.edges
                            and Jac.check_connect_access(d)
                        ):
                            if not d.edges and not isinstance(d.archetype, Root):
                                self.__shelf__.pop(_id, None)
                                continue
                            p_d.edges = d.edges

                        if Jac.check_write_access(d):
                            if hash(dumps(p_d.access)) != hash(dumps(d.access)):
                                p_d.access = d.access
                            if hash(dumps(p_d.archetype)) != hash(dumps(d.archetype)):
                                p_d.archetype = d.archetype

                        self.__shelf__[_id] = p_d
                    elif not (
                        isinstance(d, NodeAnchor)
                        and not isinstance(d.archetype, Root)
                        and not d.edges
                    ):
                        self.__shelf__[_id] = d

    def query(
        self, filter: Callable[[Anchor], bool] | None = None
    ) -> Generator[Any, None, None]:
        """Find anchors from memory with filter."""
        if isinstance(self.__shelf__, Shelf):
            for anchor in self.__shelf__.values():
                if not filter or filter(anchor):
                    if anchor.id not in self.__mem__:
                        self.__mem__[anchor.id] = anchor
                    yield anchor
        else:
            yield from super().query(filter)

    def find(
        self,
        ids: UUID | Iterable[UUID],
        filter: Callable[[Anchor], Anchor] | None = None,
    ) -> Generator[Anchor, None, None]:
        """Find anchors from datasource by ids with filter."""
        if not isinstance(ids, Iterable):
            ids = [ids]

        if isinstance(self.__shelf__, Shelf):
            for id in ids:
                anchor = self.__mem__.get(id)

                if (
                    not anchor
                    and id not in self.__gc__
                    and (_anchor := self.__shelf__.get(str(id)))
                ):
                    self.__mem__[id] = anchor = _anchor
                if anchor and (not filter or filter(anchor)):
                    yield anchor
        else:
            yield from super().find(ids, filter)

    def find_by_id(self, id: UUID) -> Anchor | None:
        """Find one by id."""
        data = super().find_by_id(id)

        if (
            not data
            and isinstance(self.__shelf__, Shelf)
            and (data := self.__shelf__.get(str(id)))
        ):
            self.__mem__[id] = data

        return data



"""Tests for Jac parser."""

import inspect
from typing import List, Type

from jaclang.runtimelib.machine import (
    JacMachineInterface,
    JacMachineImpl,
    JacMachineSpec,
)
from jaclang.utils.test import TestCase

import pluggy


class TestFeatures(TestCase):
    """Test Jac self.prse."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def get_methods(self, cls: Type) -> List[str]:
        """Get a list of method names with their signatures for a given class."""
        methods = []
        for name, value in inspect.getmembers(cls, predicate=inspect.isfunction):
            value = getattr(cls, name)
            # Get the signature using inspect
            signature = inspect.signature(value)
            new_parameters = []
            for (
                _,
                param,
            ) in signature.parameters.items():  # to strip defaults
                new_param = param.replace(default=inspect.Parameter.empty)
                new_parameters.append(new_param)
            signature = signature.replace(parameters=new_parameters)
            methods.append(f"{name}{signature}")
        return methods

    def test_feature_funcs_synced(self) -> None:
        """Test if JacMachine, JacMachineDefaults, and JacMachineSpec have synced methods."""
        # Get methods of each class
        jac_feature_methods = self.get_methods(JacMachineInterface)
        jac_feature_defaults_methods = self.get_methods(JacMachineImpl)
        jac_feature_spec_methods = self.get_methods(JacMachineSpec)

        # Check if all methods are the same in all classes
        self.assertGreater(len(jac_feature_methods), 5)
        self.assertEqual(jac_feature_spec_methods, jac_feature_defaults_methods)
        for i in jac_feature_spec_methods:
            self.assertIn(i, jac_feature_methods)

    def test_multiple_plugins(self) -> None:
        """Test that multiple plugins can implement the same hook."""
        pm = pluggy.PluginManager("jac")
        hookimpl = pluggy.HookimplMarker("jac")

        class AnotherPlugin:
            @staticmethod
            @hookimpl
            def setup() -> None:
                return "I'm here"

        pm.register(AnotherPlugin())

        # Check that both implementations are detected
        self.assertEqual(len(pm.hook.setup.get_hookimpls()), 1)

        # Execute the hook and check both results are returned
        results = pm.hook.setup()
        self.assertIn("I'm here", results)



"""Unit test utilities for JacLangServer."""

import os
import tempfile

from jaclang.vendor.pygls.uris import from_fs_path
from jaclang.vendor.pygls.workspace import Workspace

from textwrap import dedent


def get_jac_file_path():
    """Return the absolute path to the sample Jac file used for testing."""
    return os.path.abspath(
        os.path.join(
            os.path.dirname(__file__), "../../../../examples/manual_code/circle.jac"
        )
    )


def create_temp_jac_file(initial_content: str = "") -> str:
    """Create a temporary Jac file with optional initial content and return its path."""
    temp = tempfile.NamedTemporaryFile(
        delete=False, suffix=".jac", mode="w", encoding="utf-8"
    )
    temp.write(initial_content)
    temp.close()
    return temp.name


def get_code(code: str) -> str:
    """Generate a sample Jac code snippet with optional test code injected."""
    jac_code = dedent(
        f'''
    """
    This module demonstrates a simple circle class and a function to calculate
    the area of a circle in all of Jac's glory.
    """

    import math;

    # Module-level global variable
    glob RAD = 5;

    """Function to calculate the area of a circle."""
    def calculate_area(radius: float) -> float {{
        return math.pi * radius * radius;
    }}

    #* (This is a multiline comment in Jac)
    Above we have the demonstration of a function to calculate the area of a circle.
    Below we have the demonstration of a class to calculate the area of a circle.
    *#

    """Enum for shape types"""
    enum ShapeType {{
        CIRCLE = "Circle",
        UNKNOWN = "Unknown"
    }}

    """Base class for a shape."""
    obj Shape {{
        has shape_type: ShapeType;

        """Abstract method to calculate the area of a shape."""
        def area -> float abs;
    }}

    """Circle class inherits from Shape."""
    obj Circle(Shape) {{
        def init(radius: float) {{
            super.init(ShapeType.CIRCLE);
            self.radius = radius;
        }}

        """Overridden method to calculate the area of the circle."""
        override def area -> float {{
            return math.pi * self.radius * self.radius;
        }}
    }}

    with entry {{
        c = Circle(RAD);
    }}

    # Global also works here

    with entry:__main__ {{
        # To run the program functionality
        print(
            f"Area of a circle with radius {{RAD}} using function: {{calculate_area(RAD)}}"
        );
        print(
            f"Area of a {{c.shape_type.value}} with radius {{RAD}} using class: {{c.area()}}"
        );
    }}

    # Unit Tests!
    glob expected_area = 78.53981633974483;
    {code}

    test calc_area {{
        check almostEqual(calculate_area(RAD), expected_area);
    }}

    test circle_area {{
        c = Circle(RAD);
        check almostEqual(c.area(), expected_area);
    }}

    test circle_type {{
        c = Circle(RAD);
        check c.shape_type == ShapeType.CIRCLE;
    }}
'''
    )
    return jac_code


def get_simple_code(code: str) -> str:
    """Generate a sample Jac code snippet with optional test code injected."""
    jac_code = dedent(
        f"""
    
    # Unit Tests!
    glob expected_area0 = 78.53981633974483;
    glob expected_area1 = 78.53981633974483;
    {code}
    glob expected_area2 = 78.53981633974483;


"""
    )
    return jac_code


def create_ls_with_workspace(file_path: str):
    """Create JacLangServer and workspace for a given file path, return (uri, ls)."""
    from jaclang.langserve.engine import JacLangServer

    ls = JacLangServer()
    uri = from_fs_path(file_path)
    ls.lsp._workspace = Workspace(os.path.dirname(file_path), ls)
    return uri, ls



"""Test pass module."""

import io
import re
import sys
import unittest

import jaclang.compiler.unitree as uni
from jaclang.cli import cli
from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCase


class ImportPassPassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_pygen_jac_cli(self) -> None:
        """Basic test for pass."""
        (out := JacProgram()).build(self.fixture_abs_path("base.jac"))
        self.assertFalse(out.errors_had)
        mod = out.mod.hub[self.fixture_abs_path("impl/imps.jac")]
        self.assertIn("56", str(mod.to_dict()))

    def test_import_auto_impl(self) -> None:
        """Basic test for pass."""
        (prog := JacProgram()).compile(self.fixture_abs_path("autoimpl.jac"))
        num_modules = len(list(prog.mod.hub.values())[0].impl_mod)
        mod_names = [i.name for i in list(prog.mod.hub.values())[0].impl_mod]
        self.assertEqual(num_modules, 4)
        self.assertIn("getme.impl", mod_names)
        self.assertIn("autoimpl.impl", mod_names)
        self.assertIn("autoimpl.something.else.impl", mod_names)

    def test_import_include_auto_impl(self) -> None:
        """Basic test for pass."""
        (prog := JacProgram()).build(self.fixture_abs_path("incautoimpl.jac"))
        num_modules = len(list(prog.mod.hub.values())[1].impl_mod) + 1
        mod_names = [i.name for i in list(prog.mod.hub.values())[1].impl_mod]
        self.assertEqual(num_modules, 5)
        self.assertEqual("incautoimpl", list(prog.mod.hub.values())[0].name)
        self.assertEqual("autoimpl", list(prog.mod.hub.values())[1].name)
        self.assertIn("getme.impl", mod_names)
        self.assertIn("autoimpl.impl", mod_names)
        self.assertIn("autoimpl.something.else.impl", mod_names)

    def test_annexalbe_by_discovery(self) -> None:
        """Basic test for pass."""
        (prog := JacProgram()).build(self.fixture_abs_path("incautoimpl.jac"))
        count = 0
        all_mods = prog.mod.hub.values()
        self.assertEqual(len(all_mods), 6)
        for main_mod in all_mods:
            for i in main_mod.impl_mod:
                if i.name not in ["autoimpl", "incautoimpl"]:
                    count += 1
                    self.assertEqual(
                        i.annexable_by, self.fixture_abs_path("autoimpl.jac")
                    )
        self.assertEqual(count, 4)

    @unittest.skip("TODO: Fix when we have the type checker")
    def test_py_raise_map(self) -> None:
        """Basic test for pass."""
        (build := JacProgram()).build(self.fixture_abs_path("py_imp_test.jac"))
        p = {
            "math": r"jaclang/vendor/mypy/typeshed/stdlib/math.pyi$",
            "pygame_mock": r"pygame_mock/__init__.pyi$",
            "pygame_mock.color": r"pygame_mock/color.py$",
            "pygame_mock.constants": r"pygame_mock/constants.py$",
            "argparse": r"jaclang/vendor/mypy/typeshed/stdlib/argparse.pyi$",
            "builtins": r"jaclang/vendor/mypy/typeshed/stdlib/builtins.pyi$",
            "pygame_mock.display": r"pygame_mock/display.py$",
            "os": r"jaclang/vendor/mypy/typeshed/stdlib/os/__init__.pyi$",
            "genericpath": r"jaclang/vendor/mypy/typeshed/stdlib/genericpath.pyi$",
        }
        for i in p:
            self.assertIn(i, build.py_raise_map)
            self.assertRegex(
                re.sub(r".*fixtures/", "", build.py_raise_map[i]).replace("\\", "/"),
                p[i],
            )

    @unittest.skip("TODO: Fix when we have the type checker")
    def test_py_raised_mods(self) -> None:
        """Basic test for pass."""
        (prog := JacProgram()).build(self.fixture_abs_path("py_imp_test.jac"))
        for i in list(
            filter(
                lambda x: x.is_raised_from_py,
                prog.mod.hub.values(),
            )
        ):
            print(uni.Module.get_href_path(i))

        module_count = len(
            list(
                filter(
                    lambda x: x.is_raised_from_py,
                    prog.mod.hub.values(),
                )
            )
        )

        self.assertEqual(module_count, 8)

    def test_double_empty_anx(self) -> None:
        """Test importing python."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.run(self.fixture_abs_path("autoimpl.jac"))
        cli.run(self.fixture_abs_path("autoimpl.jac"))
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("foo", stdout_value)
        self.assertIn("bar", stdout_value)
        self.assertIn("baz", stdout_value)

    def test_circular_import(self) -> None:
        """Test circular import."""
        (state := JacProgram()).compile(self.fixture_abs_path("circular_import.jac"))
        self.assertFalse(state.errors_had)
        self.assertEqual(len(state.errors_had), 0)



# flake8: noqa

"""Python to Jac convertion test."""

from dataclasses import dataclass


@dataclass
class Inner:
    x: int
    y: int


@dataclass
class Container:
    inner: Inner


a = 1
b = 2

match Container(inner=Inner(x=a, y=b)):
    case Container(inner=Inner(x=a, y=0)):
        print(f"1.Inner.x={a}, Inner.y=0")
    case Container(inner=Inner(x=0, y=b)):
        print(f"2.Inner.x=0, Inner.y={b}")
    case Container(inner=Inner(x=a, y=b)):
        print(f"4.Inner.x={a}, Inner.y={b}")
    case _:
        print("5.No match")



"""Annex module loading pass for the Jac compiler.

This pass handles the discovery, loading, and attachment of annex modules to their base modules.
Annex modules are specialized extension files (.impl.jac and .test.jac) that provide
implementations or tests for a base module. The pass:

1. Identifies annex files related to the current module
2. Compiles these annex modules
3. Attaches them to the appropriate base module
4. Maintains proper relationships between base modules and their annexes

This enables the separation of interface and implementation, as well as test code organization.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING

from jaclang.compiler import unitree as uni
from jaclang.compiler.passes import Transform
from jaclang.settings import settings

if TYPE_CHECKING:
    from jaclang.compiler.program import JacProgram


class JacAnnexPass(Transform[uni.Module, uni.Module]):
    """Handles loading and attaching of annex files (.impl.jac and .test.jac)."""

    def transform(self, ir_in: uni.Module) -> uni.Module:
        """Initialize JacAnnexPass with the module path."""
        self.mod_path = ir_in.loc.mod_path
        self.base_path = self.mod_path[:-4]
        self.impl_folder = self.base_path + ".impl"
        self.test_folder = self.base_path + ".test"
        self.directory = os.path.dirname(self.mod_path) or os.getcwd()
        self.load_annexes(jac_program=self.prog, node=ir_in)
        return ir_in

    def find_annex_paths(self) -> list[str]:
        """Return list of .impl.jac and .test.jac files related to base module."""
        paths = [os.path.join(self.directory, f) for f in os.listdir(self.directory)]
        for folder in [self.impl_folder, self.test_folder]:
            if os.path.exists(folder):
                paths += [os.path.join(folder, f) for f in os.listdir(folder)]
        return paths

    def load_annexes(self, jac_program: JacProgram, node: uni.Module) -> None:
        """Parse and attach annex modules to the node."""
        if node.stub_only or not self.mod_path.endswith(".jac"):
            return
        if not self.mod_path:
            self.log_error("Module path is empty.")
            return

        for path in self.find_annex_paths():
            if path == self.mod_path:
                continue

            if path.endswith(".impl.jac") and (
                path.startswith(f"{self.base_path}.")
                or os.path.dirname(path) == self.impl_folder
            ):
                mod = jac_program.compile(file_path=path, no_cgen=True)
                if mod:
                    node.impl_mod.append(mod)

            elif (
                path.endswith(".test.jac")
                and not settings.ignore_test_annex
                and (
                    path.startswith(f"{self.base_path}.")
                    or os.path.dirname(path) == self.test_folder
                )
            ):
                mod = jac_program.compile(file_path=path, no_cgen=True)
                if mod:
                    node.test_mod.append(mod)



"""The Jac Programming Language."""

import sys

from jaclang.runtimelib.machine import (
    JacMachine,
    JacMachineImpl,
    JacMachineInterface,
    plugin_manager,
)
from jaclang.runtimelib.meta_importer import JacMetaImporter


plugin_manager.register(JacMachineImpl)
plugin_manager.load_setuptools_entrypoints("jac")

if not any(isinstance(f, JacMetaImporter) for f in sys.meta_path):
    sys.meta_path.insert(0, JacMetaImporter())

__all__ = ["JacMachineInterface", "JacMachine"]



# flake8: noqa
from __future__ import annotations

def bool_expr(prev_token_index, next_token_index, tok, change_end_line, change_end_char):
    if (not (prev_token_index is None or next_token_index is None)) and (tok[0] > change_end_line or (tok[0] == change_end_line and tok[1] > change_end_char)):
        return True
    return False



"""Test pass module."""

from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCase


class DefUsePassTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_def_uses(self) -> None:
        """Basic test for pass."""
        state = JacProgram().compile(
            file_path=self.fixture_abs_path("defs_and_uses.jac")
        )
        uses = [i.uses for i in state.sym_tab.kid_scope[0].names_in_scope.values()]
        self.assertEqual(len(uses[0]), 1)
        self.assertEqual(len(uses[1]), 1)
        self.assertIn("output", [uses[0][0].sym_name, uses[1][0].sym_name])
        self.assertIn("message", [uses[0][0].sym_name, uses[1][0].sym_name])



"""Core constructs for Jac Language."""

from __future__ import annotations

from dataclasses import asdict, dataclass, field, fields, is_dataclass
from enum import IntEnum
from functools import cached_property
from inspect import _empty, signature
from logging import getLogger
from pickle import dumps
from types import UnionType
from typing import Any, Callable, ClassVar, Optional, TypeAlias, TypeVar
from uuid import UUID, uuid4

from ..compiler.constant import EdgeDir


logger = getLogger(__name__)

TARCH = TypeVar("TARCH", bound="Archetype")
TANCH = TypeVar("TANCH", bound="Anchor")
T = TypeVar("T")


class AccessLevel(IntEnum):
    """Access level enum."""

    NO_ACCESS = -1
    READ = 0
    CONNECT = 1
    WRITE = 2

    @staticmethod
    def cast(val: int | str | AccessLevel) -> AccessLevel:
        """Cast access level."""
        match val:
            case int():
                return AccessLevel(val)
            case str():
                return AccessLevel[val.upper()]
            case _:
                return val


@dataclass
class Access:
    """Access Structure."""

    anchors: dict[str, AccessLevel] = field(default_factory=dict)

    def check(self, anchor: str) -> AccessLevel | None:
        """Validate access."""
        return self.anchors.get(anchor)


@dataclass
class Permission:
    """Anchor Access Handler."""

    all: AccessLevel = AccessLevel.NO_ACCESS
    roots: Access = field(default_factory=Access)


@dataclass
class AnchorReport:
    """Report Handler."""

    id: str
    context: dict[str, Any]


DataSpatialFilter: TypeAlias = (
    Callable[["Archetype"], bool] | "Archetype" | list["Archetype"] | None
)


@dataclass(eq=False, repr=False)
class DataSpatialDestination:
    """Object-Spatial Destination."""

    direction: EdgeDir
    edge: Callable[["Archetype"], bool] | None = None
    node: Callable[["Archetype"], bool] | None = None

    def edge_filter(self, arch: Archetype) -> bool:
        """Filter edge."""
        return not self.edge or self.edge(arch)

    def node_filter(self, arch: Archetype) -> bool:
        """Filter node."""
        return not self.node or self.node(arch)


@dataclass(eq=False, repr=False)
class DataSpatialPath:
    """Object-Spatial Path."""

    origin: list[NodeArchetype]
    destinations: list[DataSpatialDestination]
    edge_only: bool
    from_visit: bool

    def __init__(
        self,
        origin: NodeArchetype | list[NodeArchetype],
        destinations: list[DataSpatialDestination] | None = None,
    ) -> None:
        """Override Init."""
        if not isinstance(origin, list):
            origin = [origin]
        self.origin = origin
        self.destinations = [] if destinations is None else destinations
        self.edge_only = False
        self.from_visit = False

    def convert(
        self,
        filter: DataSpatialFilter,
    ) -> Callable[["Archetype"], bool] | None:
        """Convert filter."""
        if not filter:
            return None
        if callable(filter):
            return filter
        elif isinstance(filter, list):
            return lambda i: i in filter
        return lambda i: i == filter

    def append(
        self,
        direction: EdgeDir,
        edge: DataSpatialFilter,
        node: DataSpatialFilter,
    ) -> DataSpatialPath:
        """Append destination."""
        self.destinations.append(
            DataSpatialDestination(direction, self.convert(edge), self.convert(node))
        )
        return self

    def _out(
        self, edge: DataSpatialFilter = None, node: DataSpatialFilter = None
    ) -> DataSpatialPath:
        """Override greater than function."""
        return self.append(EdgeDir.OUT, edge, node)

    def _in(
        self, edge: DataSpatialFilter = None, node: DataSpatialFilter = None
    ) -> DataSpatialPath:
        """Override greater than function."""
        return self.append(EdgeDir.IN, edge, node)

    def _any(
        self, edge: DataSpatialFilter = None, node: DataSpatialFilter = None
    ) -> DataSpatialPath:
        """Override greater than function."""
        return self.append(EdgeDir.ANY, edge, node)

    def edge(self) -> DataSpatialPath:
        """Set edge only."""
        self.edge_only = True
        return self

    def visit(self) -> DataSpatialPath:
        """Set from visit."""
        self.from_visit = True
        return self

    def repr_builder(self, repr: str, dest: DataSpatialDestination, mark: str) -> str:
        """Repr builder."""
        repr += mark
        repr += f' (edge{" filter" if dest.edge else ""}) '
        repr += mark
        repr += f' (node{" filter" if dest.node else ""}) '
        return repr

    def __repr__(self) -> str:
        """Override repr."""
        repr = "nodes "
        for dest in self.destinations:
            match dest.direction:
                case EdgeDir.IN:
                    repr = self.repr_builder(repr, dest, "<<")
                case EdgeDir.OUT:
                    repr = self.repr_builder(repr, dest, ">>")
                case _:
                    repr = self.repr_builder(repr, dest, "--")
        return repr.strip()


@dataclass(eq=False, repr=False, kw_only=True)
class Anchor:
    """Object Anchor."""

    archetype: Archetype
    id: UUID = field(default_factory=uuid4)
    root: Optional[UUID] = None
    access: Permission = field(default_factory=Permission)
    persistent: bool = False
    hash: int = 0

    def is_populated(self) -> bool:
        """Check if state."""
        return "archetype" in self.__dict__

    def make_stub(self: TANCH) -> TANCH:
        """Return unsynced copy of anchor."""
        if self.is_populated():
            unloaded = object.__new__(self.__class__)
            unloaded.id = self.id
            return unloaded
        return self

    def populate(self) -> None:
        """Retrieve the Archetype from db and return."""
        from jaclang.runtimelib.machine import JacMachineInterface as Jac

        jsrc = Jac.get_context().mem

        if anchor := jsrc.find_by_id(self.id):
            self.__dict__.update(anchor.__dict__)

    def __getattr__(self, name: str) -> object:
        """Trigger load if detects unloaded state."""
        if not self.is_populated():
            self.populate()

            if not self.is_populated():
                raise ValueError(
                    f"{self.__class__.__name__} [{self.id}] is not a valid reference!"
                )

            return getattr(self, name)

        raise AttributeError(
            f"'{self.__class__.__name__}' object has not attribute '{name}'"
        )

    def __getstate__(self) -> dict[str, Any]:  # NOTE: May be better type hinting
        """Serialize Anchor."""
        if self.is_populated():
            unlinked = object.__new__(self.archetype.__class__)
            unlinked.__dict__.update(self.archetype.__dict__)
            unlinked.__dict__.pop("__jac__", None)

            return {
                "id": self.id,
                "archetype": unlinked,
                "root": self.root,
                "access": self.access,
                "persistent": self.persistent,
            }
        else:
            return {"id": self.id}

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Deserialize Anchor."""
        self.__dict__.update(state)

        if self.is_populated() and self.archetype:
            self.archetype.__jac__ = self
            self.hash = hash(dumps(self))

    def __repr__(self) -> str:
        """Override representation."""
        if self.is_populated():
            attrs = ""
            for f in fields(self):
                if f.name in self.__dict__:
                    attrs += f"{f.name}={self.__dict__[f.name]}, "
            attrs = attrs[:-2]
        else:
            attrs = f"id={self.id}"

        return f"{self.__class__.__name__}({attrs})"

    def report(self) -> AnchorReport:
        """Report Anchor."""
        return AnchorReport(
            id=self.id.hex,
            context=(
                asdict(self.archetype)
                if is_dataclass(self.archetype) and not isinstance(self.archetype, type)
                else {}
            ),
        )

    def __hash__(self) -> int:
        """Override hash for anchor."""
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Override equal implementation."""
        if isinstance(other, Anchor):
            return self.__class__ is other.__class__ and self.id == other.id

        return False


@dataclass(eq=False, repr=False, kw_only=True)
class NodeAnchor(Anchor):
    """Node Anchor."""

    archetype: NodeArchetype
    edges: list[EdgeAnchor]

    def __getstate__(self) -> dict[str, object]:
        """Serialize Node Anchor."""
        state = super().__getstate__()

        if self.is_populated():
            state["edges"] = [edge.make_stub() for edge in self.edges]

        return state


@dataclass(eq=False, repr=False, kw_only=True)
class EdgeAnchor(Anchor):
    """Edge Anchor."""

    archetype: EdgeArchetype
    source: NodeAnchor
    target: NodeAnchor
    is_undirected: bool

    def __getstate__(self) -> dict[str, object]:
        """Serialize Node Anchor."""
        state = super().__getstate__()

        if self.is_populated():
            state.update(
                {
                    "source": self.source.make_stub(),
                    "target": self.target.make_stub(),
                    "is_undirected": self.is_undirected,
                }
            )

        return state


@dataclass(eq=False, repr=False, kw_only=True)
class WalkerAnchor(Anchor):
    """Walker Anchor."""

    archetype: WalkerArchetype
    path: list[NodeAnchor] = field(default_factory=list)
    next: list[NodeAnchor | EdgeAnchor] = field(default_factory=list)
    ignores: list[NodeAnchor] = field(default_factory=list)
    disengaged: bool = False


@dataclass(eq=False, repr=False, kw_only=True)
class ObjectAnchor(Anchor):
    """Edge Anchor."""

    archetype: ObjectArchetype


@dataclass(eq=False, repr=False, kw_only=True)
class Archetype:
    """Archetype Protocol."""

    _jac_entry_funcs_: ClassVar[list[DataSpatialFunction]] = []
    _jac_exit_funcs_: ClassVar[list[DataSpatialFunction]] = []

    @cached_property
    def __jac__(self) -> Anchor:
        """Create default anchor."""
        return Anchor(archetype=self)

    def __init_subclass__(cls) -> None:
        """Configure subclasses."""
        if not cls.__dict__.get("__jac_base__", False):
            from jaclang import JacMachineInterface as _

            _.make_archetype(cls)

    def __repr__(self) -> str:
        """Override repr for archetype."""
        return f"{self.__class__.__name__}"

    def __jac_access__(self) -> AccessLevel | str | int | None:
        """Override access validation."""
        return None


class NodeArchetype(Archetype):
    """Node Archetype Protocol."""

    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> NodeAnchor:
        """Create default anchor."""
        return NodeAnchor(archetype=self, edges=[])


class EdgeArchetype(Archetype):
    """Edge Archetype Protocol."""

    __jac_base__: ClassVar[bool] = True
    __jac__: EdgeAnchor


class WalkerArchetype(Archetype):
    """Walker Archetype Protocol."""

    __jac_async__: ClassVar[bool] = False
    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> WalkerAnchor:
        """Create default anchor."""
        return WalkerAnchor(archetype=self)


class ObjectArchetype(Archetype):
    """Walker Archetype Protocol."""

    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> ObjectAnchor:
        """Create default anchor."""
        return ObjectAnchor(archetype=self)


@dataclass(eq=False)
class GenericEdge(EdgeArchetype):
    """Generic Edge."""

    __jac_base__: ClassVar[bool] = True

    def __repr__(self) -> str:
        """Override repr for archetype."""
        return f"{self.__class__.__name__}()"


@dataclass(eq=False)
class Root(NodeArchetype):
    """Generic Root Node."""

    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> NodeAnchor:
        """Create default anchor."""
        return NodeAnchor(archetype=self, persistent=True, edges=[])

    def __repr__(self) -> str:
        """Override repr for archetype."""
        return f"{self.__class__.__name__}()"


@dataclass(eq=False)
class DataSpatialFunction:
    """Object-Spatial Function."""

    name: str
    func: Callable[[Any, Any], Any]

    @cached_property
    def trigger(self) -> type | UnionType | tuple[type | UnionType, ...] | None:
        """Get function parameter annotations."""
        parameters = signature(self.func, eval_str=True).parameters
        if len(parameters) >= 2:
            second_param = list(parameters.values())[1]
            ty = second_param.annotation
            return ty if ty != _empty else None
        return None



"""Module resolver utilities."""

from __future__ import annotations

import os
import site
import sys
from typing import Optional, Tuple


def get_jac_search_paths(base_path: Optional[str] = None) -> list[str]:
    """Construct a list of paths to search for Jac modules."""
    paths = []
    if base_path:
        paths.append(base_path)
    paths.append(os.getcwd())
    if "JACPATH" in os.environ:
        paths.extend(
            p.strip() for p in os.environ["JACPATH"].split(os.pathsep) if p.strip()
        )
    paths.extend(sys.path)
    site_pkgs = site.getsitepackages()
    if site_pkgs:
        paths.extend(site_pkgs)
    user_site = getattr(site, "getusersitepackages", None)
    if user_site:
        user_dir = user_site()
        if user_dir:
            paths.append(user_dir)
    return list(dict.fromkeys(filter(None, paths)))


def _candidate_from(base: str, parts: list[str]) -> Optional[Tuple[str, str]]:
    candidate = os.path.join(base, *parts)
    if os.path.isdir(candidate):
        if os.path.isfile(os.path.join(candidate, "__init__.jac")):
            return os.path.join(candidate, "__init__.jac"), "jac"
        if os.path.isfile(os.path.join(candidate, "__init__.py")):
            return os.path.join(candidate, "__init__.py"), "py"
    if os.path.isfile(candidate + ".jac"):
        return candidate + ".jac", "jac"
    if os.path.isfile(candidate + ".py"):
        return candidate + ".py", "py"
    return None


def resolve_module(target: str, base_path: str) -> Tuple[str, str]:
    """Resolve module path and infer language."""
    parts = target.split(".")
    level = 0
    while level < len(parts) and parts[level] == "":
        level += 1
    actual_parts = parts[level:]

    for sp in site.getsitepackages():
        res = _candidate_from(sp, actual_parts)
        if res:
            return res

    base_dir = base_path if os.path.isdir(base_path) else os.path.dirname(base_path)
    for _ in range(max(level - 1, 0)):
        base_dir = os.path.dirname(base_dir)
    res = _candidate_from(base_dir, actual_parts)
    if res:
        return res

    jacpath = os.getenv("JACPATH")
    if jacpath:
        res = _candidate_from(jacpath, actual_parts)
        if res:
            return res
        target_jac = actual_parts[-1] + ".jac"
        target_py = actual_parts[-1] + ".py"
        for root, _, files in os.walk(jacpath):
            if target_jac in files:
                return os.path.join(root, target_jac), "jac"
            if target_py in files:
                return os.path.join(root, target_py), "py"

    return os.path.join(base_dir, *actual_parts), "py"


def infer_language(target: str, base_path: str) -> str:
    """Infer language for target relative to base path."""
    _, lang = resolve_module(target, base_path)
    return lang


def resolve_relative_path(target: str, base_path: str) -> str:
    """Resolve only the path component for a target."""
    path, _ = resolve_module(target, base_path)
    return path



"""Test Jac cli module."""

import contextlib
import inspect
import io
import os
import re
import subprocess
import sys
import traceback
import unittest
from jaclang.cli import cli
from jaclang.cli.cmdreg import cmd_registry, extract_param_descriptions
from jaclang.runtimelib.builtin import printgraph
from jaclang.utils.test import TestCase


class JacCliTests(TestCase):
    """Test pass module."""

    def setUp(self) -> None:
        """Set up test."""
        return super().setUp()

    def test_jac_cli_run(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        # Execute the function
        cli.run(self.fixture_abs_path("hello.jac"))

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()

        self.assertIn("Hello World!", stdout_value)

    def test_jac_cli_alert_based_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        try:
            cli.enter(self.fixture_abs_path("err2.jac"), entrypoint="speak", args=[])
        except Exception as e:
            print(f"Error: {e}")

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        # print(stdout_value)
        self.assertIn("Error", stdout_value)

    def test_jac_cli_alert_based_runtime_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        with contextlib.suppress(Exception):
            cli.run(self.fixture_abs_path("err_runtime.jac"))

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        expected_stdout_values = (
            "Error: list index out of range",
            "    print(some_list[invalid_index]);",
            "          ^^^^^^^^^^^^^^^^^^^^^^^^",
            "  at bar() ",
            "  at foo() ",
            "  at <module> ",
        )
        logger_capture = "\n".join([rec.message for rec in self.caplog.records])
        for exp in expected_stdout_values:
            self.assertIn(exp, logger_capture)

    def test_jac_impl_err(self) -> None:
        """Basic test for pass."""
        if "jaclang.tests.fixtures.err" in sys.modules:
            del sys.modules["jaclang.tests.fixtures.err"]
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output

        try:
            cli.enter(self.fixture_abs_path("err.jac"), entrypoint="speak", args=[])
        except Exception:
            traceback.print_exc()

        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        # print(stdout_value)
        path_to_file = self.fixture_abs_path("err.impl.jac")
        self.assertIn(f'"{path_to_file}", line 2', stdout_value)

    def test_param_name_diff(self) -> None:
        """Test when parameter name from definitinon and declaration are mismatched."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output
        with contextlib.suppress(Exception):
            cli.run(self.fixture_abs_path("decl_defn_param_name.jac"))
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        expected_stdout_values = (
            "short_name = 42",
            "p1 = 64 , p2 = foobar",
        )
        output = captured_output.getvalue()
        for exp in expected_stdout_values:
            self.assertIn(exp, output)

    def test_jac_test_err(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        sys.stderr = captured_output
        cli.test(self.fixture_abs_path("baddy.jac"))
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        stdout_value = captured_output.getvalue()
        path_to_file = self.fixture_abs_path("baddy.test.jac")
        self.assertIn(f'"{path_to_file}", line 2,', stdout_value)

    def test_jac_ast_tool_pass_template(self) -> None:
        """Basic test for pass."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("pass_template")

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Sub objects.", stdout_value)
        self.assertGreater(stdout_value.count("def exit_"), 10)

    def test_jac_cmd_line(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout_value, _ = process.communicate(input="exit\n")
        self.assertEqual(process.returncode, 0, "Process did not exit successfully")
        self.assertIn("Welcome to the Jac CLI!", stdout_value)

    def test_ast_print(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('hello.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("+-- Token", stdout_value)

    @unittest.skip("Skipping builtins loading test")
    def test_builtins_loading(self) -> None:
        """Testing for print AstTool."""
        from jaclang.settings import settings

        settings.ast_symbol_info_detailed = True
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast", f"{self.fixture_abs_path('builtins_test.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        settings.ast_symbol_info_detailed = False

        self.assertRegex(
            stdout_value,
            r"2\:8 \- 2\:12.*BuiltinType - list - .*SymbolPath: builtins.list",
        )
        self.assertRegex(
            stdout_value,
            r"15\:5 \- 15\:8.*Name - dir - .*SymbolPath: builtins.dir",
        )
        self.assertRegex(
            stdout_value,
            r"13\:12 \- 13\:18.*Name - append - .*SymbolPath: builtins.list.append",
        )

    def test_ast_printgraph(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output

        cli.tool("ir", ["ast.", f"{self.fixture_abs_path('hello.jac')}"])

        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn(
            '[label="MultiString" shape="oval" style="filled" fillcolor="#fccca4"]',
            stdout_value,
        )

    def test_del_clean(self) -> None:
        """Testing for print AstTool."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.check(f"{self.fixture_abs_path('del_clean.jac')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Errors: 0, Warnings: 0", stdout_value)

    def test_build_and_run(self) -> None:
        """Testing for print AstTool."""
        if os.path.exists(f"{self.fixture_abs_path('needs_import.jir')}"):
            os.remove(f"{self.fixture_abs_path('needs_import.jir')}")
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.build(f"{self.fixture_abs_path('needs_import.jac')}")
        cli.run(f"{self.fixture_abs_path('needs_import.jir')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("Errors: 0, Warnings: 0", stdout_value)
        self.assertIn("<module 'pyfunc' from", stdout_value)

    def test_run_test(self) -> None:
        """Basic test for pass."""
        process = subprocess.Popen(
            ["jac", "test", f"{self.fixture_abs_path('run_test.jac')}", "-m 2"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("Ran 3 tests", stderr)
        self.assertIn("FAILED (failures=2)", stderr)
        self.assertIn("F.F", stderr)

        process = subprocess.Popen(
            [
                "jac",
                "test",
                "-d" + f"{self.fixture_abs_path('../../../')}",
                "-f" + "circle*",
                "-x",
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("circle", stdout)
        self.assertNotIn("circle_purfe.test", stdout)
        self.assertNotIn("circle_pure.impl", stdout)

        process = subprocess.Popen(
            ["jac", "test", "-f" + "*run_test.jac", "-m 3"],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("...F", stderr)
        self.assertIn("F.F", stderr)

    def test_run_specific_test_only(self) -> None:
        """Test a specific test case."""
        process = subprocess.Popen(
            [
                "jac",
                "test",
                "-t",
                "from_2_to_10",
                self.fixture_abs_path("jactest_main.jac"),
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            text=True,
        )
        stdout, stderr = process.communicate()
        self.assertIn("Ran 1 test", stderr)
        self.assertIn("Testing fibonacci numbers from 2 to 10.", stdout)
        self.assertNotIn("Testing first 2 fibonacci numbers.", stdout)
        self.assertNotIn("This test should not run after import.", stdout)

    def test_graph_coverage(self) -> None:
        """Test for coverage of graph cmd."""
        graph_params = set(inspect.signature(cli.dot).parameters.keys())
        printgraph_params = set(inspect.signature(printgraph).parameters.keys())
        printgraph_params = printgraph_params - {
            "node",
            "file",
            "edge_type",
            "format",
        }
        printgraph_params.update({"initial", "saveto", "connection", "session"})
        self.assertTrue(printgraph_params.issubset(graph_params))
        self.assertEqual(len(printgraph_params) + 1, len(graph_params))

    def test_graph(self) -> None:
        """Test for graph CLI cmd."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.dot(f"{self.examples_abs_path('reference/connect_expressions.jac')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        if os.path.exists("connect_expressions.dot"):
            os.remove("connect_expressions.dot")
        self.assertIn("11\n13\n15\n>>> Graph content saved to", stdout_value)
        self.assertIn("connect_expressions.dot\n", stdout_value)

    def test_py_to_jac(self) -> None:
        """Test for graph CLI cmd."""
        captured_output = io.StringIO()
        sys.stdout = captured_output
        cli.py2jac(f"{self.fixture_abs_path('../../tests/fixtures/pyfunc.py')}")
        sys.stdout = sys.__stdout__
        stdout_value = captured_output.getvalue()
        self.assertIn("def my_print(x: object) -> None", stdout_value)

    def test_caching_issue(self) -> None:
        """Test for Caching Issue."""
        test_file = self.fixture_abs_path("test_caching_issue.jac")
        test_cases = [(10, True), (11, False)]
        for x, is_passed in test_cases:
            with open(test_file, "w") as f:
                f.write(
                    f"""
                test mytest{{
                    check 10 == {x};
                }}
                """
                )
            process = subprocess.Popen(
                ["jac", "test", test_file],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            stdout, stderr = process.communicate()
            if is_passed:
                self.assertIn("Passed successfully.", stdout)
                self.assertIn(".", stderr)
            else:
                self.assertNotIn("Passed successfully.", stdout)
                self.assertIn("F", stderr)
        os.remove(test_file)

    def test_cli_docstring_parameters(self) -> None:
        """Test that all CLI command parameters are documented in their docstrings."""
        # Get all registered CLI commands
        commands = {}
        for name, _ in cmd_registry.registry.items():
            # Skip commands that might be registered from outside cli.py
            if hasattr(cli, name):
                commands[name] = getattr(cli, name)

        missing_params = {}

        for cmd_name, cmd_func in commands.items():
            # Get function parameters from signature
            signature_params = set(inspect.signature(cmd_func).parameters.keys())

            # Parse docstring to extract documented parameters
            docstring = cmd_func.__doc__ or ""

            # Check if the docstring has an Args section
            args_match = re.search(r"Args:(.*?)(?:\n\n|\Z)", docstring, re.DOTALL)
            if not args_match:
                missing_params[cmd_name] = list(signature_params)
                continue

            args_section = args_match.group(1)

            # Extract parameter names from the Args section
            # Looking for patterns like "param_name: Description" or "param_name (type): Description"
            doc_params = set()
            for line in args_section.strip().split("\n"):
                line = line.strip()
                if not line:
                    continue

                # Match parameter name at the beginning of the line
                param_match = re.match(r"\s*([a-zA-Z0-9_]+)(?:\s*\([^)]*\))?:\s*", line)
                if param_match:
                    doc_params.add(param_match.group(1))

            # Find parameters that are in the signature but not in the docstring
            undocumented_params = signature_params - doc_params
            if undocumented_params:
                missing_params[cmd_name] = list(undocumented_params)

        # Assert that there are no missing parameters
        self.assertEqual(
            missing_params,
            {},
            f"The following CLI commands have undocumented parameters: {missing_params}",
        )

    def test_cli_help_uses_docstring_descriptions(self) -> None:
        """Test that CLI help text uses parameter descriptions from docstrings."""
        # Get a command with well-documented parameters
        test_commands = ["run", "dot", "test"]

        for cmd_name in test_commands:
            # Skip if command doesn't exist
            if not hasattr(cli, cmd_name):
                continue

            cmd_func = getattr(cli, cmd_name)
            docstring = cmd_func.__doc__ or ""

            # Extract parameter descriptions from docstring
            docstring_param_descriptions = extract_param_descriptions(docstring)

            # Skip if no parameters are documented
            if not docstring_param_descriptions:
                continue

            # Get help text for the command
            captured_output = io.StringIO()
            sys.stdout = captured_output

            # Use subprocess to get the help text to ensure we're testing the actual CLI output
            process = subprocess.Popen(
                ["jac", cmd_name, "--help"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
            )
            help_text, _ = process.communicate()

            sys.stdout = sys.__stdout__

            # Check that each documented parameter description appears in the help text
            for param_name, description in docstring_param_descriptions.items():
                # The description might be truncated or formatted differently in the help text,
                # so we'll just check for the first few words
                description_start = description.split()[:3]
                description_pattern = r"\s+".join(
                    re.escape(word) for word in description_start
                )
                # Check if the description appears in the help text
                self.assertRegex(
                    help_text,
                    description_pattern,
                    f"Parameter description for '{param_name}' not found in help text for '{cmd_name}'",
                )



"""Test for jaseci plugin."""

import io
import os
import sys

from jaclang.cli import cli
from jaclang.utils.test import TestCase

session = ""


class TestJaseciPlugin(TestCase):
    """Test jaseci plugin."""

    def setUp(self) -> None:
        """Set up test."""
        super().setUp()

    def tearDown(self) -> None:
        """Tear down test."""
        super().tearDown()
        sys.stdout = sys.__stdout__

    def _output2buffer(self) -> None:
        """Start capturing output."""
        self.capturedOutput = io.StringIO()
        sys.stdout = self.capturedOutput

    def _output2std(self) -> None:
        """Redirect output back to stdout."""
        sys.stdout = sys.__stdout__

    def _del_session(self, session: str) -> None:
        path = os.path.dirname(session)
        prefix = os.path.basename(session)
        for file in os.listdir(path):
            if file.startswith(prefix):
                os.remove(f"{path}/{file}")

    def test_walker_simple_persistent(self) -> None:
        """Test simple persistent object."""
        session = self.fixture_abs_path("test_walker_simple_persistent.session")
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="create",
            args=[],
        )
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="traverse",
            args=[],
        )
        output = self.capturedOutput.getvalue().strip()
        self.assertEqual(output, "node a\nnode b")
        self._del_session(session)

    def test_entrypoint_root(self) -> None:
        """Test entrypoint being root."""
        session = self.fixture_abs_path("test_entrypoint_root.session")
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="create",
            args=[],
        )
        obj = cli.get_object(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            id="root",
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="traverse",
            args=[],
            node=str(obj["id"]),
        )
        output = self.capturedOutput.getvalue().strip()
        self.assertEqual(output, "node a\nnode b")
        self._del_session(session)

    def test_entrypoint_non_root(self) -> None:
        """Test entrypoint being non root node."""
        session = self.fixture_abs_path("test_entrypoint_non_root.session")
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="create",
            args=[],
        )
        obj = cli.get_object(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            id="root",
            session=session,
        )
        edge_obj = cli.get_object(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            id=obj["edges"][0].id.hex,
            session=session,
        )
        a_obj = cli.get_object(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            id=edge_obj["target"].id.hex,
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="traverse",
            node=str(a_obj["id"]),
            args=[],
        )
        output = self.capturedOutput.getvalue().strip()
        self._del_session(session)
        self.assertEqual(output, "node a\nnode b")

    def test_get_edge(self) -> None:
        """Test get an edge object."""
        session = self.fixture_abs_path("test_get_edge.session")
        cli.run(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
        )
        obj = cli.get_object(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
            id="root",
        )
        self.assertEqual(len(obj["edges"]), 2)
        edge_objs = [
            cli.get_object(
                filename=self.fixture_abs_path("simple_node_connection.jac"),
                session=session,
                id=e.id.hex,
            )
            for e in obj["edges"]
        ]
        node_ids = [obj["target"].id.hex for obj in edge_objs]
        node_objs = [
            cli.get_object(
                filename=self.fixture_abs_path("simple_node_connection.jac"),
                session=session,
                id=str(n_id),
            )
            for n_id in node_ids
        ]
        self.assertEqual(len(node_objs), 2)
        self.assertEqual(
            {obj["archetype"].tag for obj in node_objs}, {"first", "second"}
        )
        self._del_session(session)

    def test_filter_on_edge_get_edge(self) -> None:
        """Test filtering on edge."""
        session = self.fixture_abs_path("test_filter_on_edge_get_edge.session")
        cli.run(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
            entrypoint="filter_on_edge_get_edge",
            args=[],
        )
        self.assertEqual(
            self.capturedOutput.getvalue().strip(), "[simple_edge(index=1)]"
        )
        self._del_session(session)

    def test_filter_on_edge_get_node(self) -> None:
        """Test filtering on edge, then get node."""
        session = self.fixture_abs_path("test_filter_on_edge_get_node.session")
        cli.run(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
            entrypoint="filter_on_edge_get_node",
            args=[],
        )
        self.assertEqual(
            self.capturedOutput.getvalue().strip(), "[simple(tag='second')]"
        )
        self._del_session(session)

    def test_filter_on_node_get_node(self) -> None:
        """Test filtering on node, then get edge."""
        session = self.fixture_abs_path("test_filter_on_node_get_node.session")
        cli.run(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
            entrypoint="filter_on_node_get_node",
            args=[],
        )
        self.assertEqual(
            self.capturedOutput.getvalue().strip(), "[simple(tag='second')]"
        )
        self._del_session(session)

    def test_filter_on_edge_visit(self) -> None:
        """Test filtering on edge, then visit."""
        session = self.fixture_abs_path("test_filter_on_edge_visit.session")
        cli.run(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
            entrypoint="filter_on_edge_visit",
            args=[],
        )
        self.assertEqual(self.capturedOutput.getvalue().strip(), "simple(tag='first')")
        self._del_session(session)

    def test_filter_on_node_visit(self) -> None:
        """Test filtering on node, then visit."""
        session = self.fixture_abs_path("test_filter_on_node_visit.session")
        cli.run(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_node_connection.jac"),
            session=session,
            entrypoint="filter_on_node_visit",
            args=[],
        )
        self.assertEqual(self.capturedOutput.getvalue().strip(), "simple(tag='first')")
        self._del_session(session)

    def test_indirect_reference_node(self) -> None:
        """Test reference node indirectly without visiting."""
        session = self.fixture_abs_path("test_indirect_reference_node.session")
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="create",
            args=[],
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("simple_persistent.jac"),
            session=session,
            entrypoint="indirect_ref",
            args=[],
        )
        # FIXME: Figure out what to do with warning.
        # self.assertEqual(
        #     self.capturedOutput.getvalue().strip(),
        #     "[b(name='node b')]\n[GenericEdge()]",
        # )
        self._del_session(session)

    def test_walker_purger(self) -> None:
        """Test simple persistent object."""
        session = self.fixture_abs_path("test_walker_purger.session")
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="populate",
            args=[],
        )
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="traverse",
            args=[],
        )
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="check",
            args=[],
        )
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="purge",
            args=[],
        )
        output = self.capturedOutput.getvalue().strip()
        self.assertEqual(
            output,
            (
                "Root()\n"
                "A(id=0)\nA(id=1)\n"
                "B(id=0)\nB(id=1)\nB(id=0)\nB(id=1)\n"
                "C(id=0)\nC(id=1)\nC(id=0)\nC(id=1)\nC(id=0)\nC(id=1)\nC(id=0)\nC(id=1)\n"
                "D(id=0)\nD(id=1)\nD(id=0)\nD(id=1)\nD(id=0)\nD(id=1)\nD(id=0)\nD(id=1)\n"
                "D(id=0)\nD(id=1)\nD(id=0)\nD(id=1)\nD(id=0)\nD(id=1)\nD(id=0)\nD(id=1)\n"
                "E(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\n"
                "E(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\n"
                "E(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\n"
                "E(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\nE(id=0)\nE(id=1)\n"
                "125\n124"
            ),
        )
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="traverse",
            args=[],
        )
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="check",
            args=[],
        )
        cli.enter(
            filename=self.fixture_abs_path("graph_purger.jac"),
            session=session,
            entrypoint="purge",
            args=[],
        )
        output = self.capturedOutput.getvalue().strip()
        self.assertEqual(output, "Root()\n1\n0")

        self._del_session(session)

    def trigger_access_validation_test(
        self, give_access_to_full_graph: bool, via_all: bool = False
    ) -> None:
        """Test different access validation."""
        self._output2buffer()

        ##############################################
        #              ALLOW READ ACCESS             #
        ##############################################

        node_1 = "" if give_access_to_full_graph else self.nodes[0]
        node_2 = "" if give_access_to_full_graph else self.nodes[1]

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="allow_other_root_access",
            args=[self.roots[1], 0, via_all],
            session=session,
            root=self.roots[0],
            node=node_1,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="allow_other_root_access",
            args=[self.roots[0], 0, via_all],
            session=session,
            root=self.roots[1],
            node=node_2,
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node",
            args=[20],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node",
            args=[10],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_target_node",
            args=[20, self.nodes[1]],
            session=session,
            root=self.roots[0],
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_target_node",
            args=[10, self.nodes[0]],
            session=session,
            root=self.roots[1],
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )
        archs = self.capturedOutput.getvalue().strip().split("\n")
        self.assertTrue(len(archs) == 2)

        # --------- NO UPDATE SHOULD HAPPEN -------- #

        self.assertEqual(archs[0], "A(val=2)")
        self.assertEqual(archs[1], "A(val=1)")

        ##############################################
        #        WITH READ ACCESS BUT ELEVATED       #
        ##############################################

        self._output2buffer()

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node_forced",
            args=[20],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node_forced",
            args=[10],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )
        archs = self.capturedOutput.getvalue().strip().split("\n")
        self.assertTrue(len(archs) == 2)

        # ---------- UPDATE SHOULD HAPPEN ---------- #

        self.assertEqual(archs[0], "A(val=20)")
        self.assertEqual(archs[1], "A(val=10)")

        # ---------- DISALLOW READ ACCESS ---------- #

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="disallow_other_root_access",
            args=[self.roots[1], via_all],
            session=session,
            root=self.roots[0],
            node=node_1,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="disallow_other_root_access",
            args=[self.roots[0], via_all],
            session=session,
            root=self.roots[1],
            node=node_2,
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )
        self.assertFalse(self.capturedOutput.getvalue().strip())

        ##############################################
        #             ALLOW WRITE ACCESS             #
        ##############################################

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="allow_other_root_access",
            args=[self.roots[1], "WRITE", via_all],
            session=session,
            root=self.roots[0],
            node=node_1,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="allow_other_root_access",
            args=[self.roots[0], "WRITE", via_all],
            session=session,
            root=self.roots[1],
            node=node_2,
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node",
            args=[200],
            root=self.roots[0],
            node=self.nodes[1],
            session=session,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node",
            args=[100],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )
        archs = self.capturedOutput.getvalue().strip().split("\n")
        self.assertTrue(len(archs) == 2)

        # ---------- UPDATE SHOULD HAPPEN ---------- #

        self.assertEqual(archs[0], "A(val=200)")
        self.assertEqual(archs[1], "A(val=100)")

        # ---------- DISALLOW WRITE ACCESS --------- #

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="disallow_other_root_access",
            args=[self.roots[1], via_all],
            session=session,
            root=self.roots[0],
            node=node_1,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="disallow_other_root_access",
            args=[self.roots[0], via_all],
            session=session,
            root=self.roots[1],
            node=node_2,
        )

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[0],
            node=self.nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=self.roots[1],
            node=self.nodes[0],
        )
        self.assertFalse(self.capturedOutput.getvalue().strip())

        # ---------- ROOTS RESET OWN NODE ---------- #

        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node",
            args=[1],
            session=session,
            root=self.roots[0],
            node=self.nodes[0],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="update_node",
            args=[2],
            session=session,
            root=self.roots[1],
            node=self.nodes[1],
        )

    def test_other_root_access(self) -> None:
        """Test filtering on node, then visit."""
        global session
        session = self.fixture_abs_path("other_root_access.session")

        ##############################################
        #                CREATE ROOTS                #
        ##############################################

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="create_other_root",
            args=[],
            session=session,
        )
        root1 = self.capturedOutput.getvalue().strip()

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="create_other_root",
            args=[],
            session=session,
        )
        root2 = self.capturedOutput.getvalue().strip()

        ##############################################
        #           CREATE RESPECTIVE NODES          #
        ##############################################

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="create_node",
            args=[1],
            session=session,
            root=root1,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="create_node",
            args=[2],
            session=session,
            root=root2,
        )
        nodes = self.capturedOutput.getvalue().strip().split("\n")
        self.assertTrue(len(nodes) == 2)

        ##############################################
        #           VISIT RESPECTIVE NODES           #
        ##############################################

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=root1,
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=root2,
        )
        archs = self.capturedOutput.getvalue().strip().split("\n")
        self.assertEqual(2, len(archs))
        self.assertTrue(archs[0], "A(val=1)")
        self.assertTrue(archs[1], "A(val=2)")

        ##############################################
        #              SWAP TARGET NODE              #
        #                  NO ACCESS                 #
        ##############################################

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=root1,
            node=nodes[1],
        )
        cli.enter(
            filename=self.fixture_abs_path("other_root_access.jac"),
            entrypoint="check_node",
            args=[],
            session=session,
            root=root2,
            node=nodes[0],
        )
        self.assertFalse(self.capturedOutput.getvalue().strip())

        ##############################################
        #        TEST DIFFERENT ACCESS OPTIONS       #
        ##############################################

        self.roots = [root1, root2]
        self.nodes = nodes

        self.trigger_access_validation_test(give_access_to_full_graph=False)
        self.trigger_access_validation_test(give_access_to_full_graph=True)

        self.trigger_access_validation_test(
            give_access_to_full_graph=False, via_all=True
        )
        self.trigger_access_validation_test(
            give_access_to_full_graph=True, via_all=True
        )

        self._del_session(session)

    def test_savable_object(self) -> None:
        """Test ObjectAnchor save."""
        global session
        session = self.fixture_abs_path("savable_object.session")

        self._output2buffer()

        cli.enter(
            filename=self.fixture_abs_path("savable_object.jac"),
            entrypoint="create_custom_object",
            args=[],
            session=session,
        )

        prints = self.capturedOutput.getvalue().strip().split("\n")
        id = prints[0]

        self.assertEqual(
            "SavableObject(val=0, arr=[], json={}, parent=Parent(val=1, arr=[1], json"
            "={'a': 1}, enum_field=<Enum.B: 'b'>, child=Child(val=2, arr=[1, 2], json"
            "={'a': 1, 'b': 2}, enum_field=<Enum.C: 'c'>)), enum_field=<Enum.A: 'a'>)",
            prints[1],
        )

        self._output2buffer()

        cli.enter(
            filename=self.fixture_abs_path("savable_object.jac"),
            entrypoint="get_custom_object",
            args=[id],
            session=session,
        )
        self.assertEqual(
            "SavableObject(val=0, arr=[], json={}, parent=Parent(val=1, arr=[1], json"
            "={'a': 1}, enum_field=<Enum.B: 'b'>, child=Child(val=2, arr=[1, 2], json"
            "={'a': 1, 'b': 2}, enum_field=<Enum.C: 'c'>)), enum_field=<Enum.A: 'a'>)",
            self.capturedOutput.getvalue().strip(),
        )

        self._output2buffer()

        cli.enter(
            filename=self.fixture_abs_path("savable_object.jac"),
            entrypoint="update_custom_object",
            args=[id],
            session=session,
        )

        self.assertEqual(
            "SavableObject(val=1, arr=[1], json={'a': 1}, parent=Parent(val=2, arr=[1, 2], json"
            "={'a': 1, 'b': 2}, enum_field=<Enum.C: 'c'>, child=Child(val=3, arr=[1, 2, 3], json"
            "={'a': 1, 'b': 2, 'c': 3}, enum_field=<Enum.A: 'a'>)), enum_field=<Enum.B: 'b'>)",
            self.capturedOutput.getvalue().strip(),
        )

        self._output2buffer()

        cli.enter(
            filename=self.fixture_abs_path("savable_object.jac"),
            entrypoint="delete_custom_object",
            args=[id],
            session=session,
        )

        cli.enter(
            filename=self.fixture_abs_path("savable_object.jac"),
            entrypoint="get_custom_object",
            args=[id],
            session=session,
        )
        self.assertEqual("None", self.capturedOutput.getvalue().strip())

        self._del_session(session)

    def test_traversing_save(self) -> None:
        """Test traversing save."""
        global session
        session = self.fixture_abs_path("traversing_save.session")

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("traversing_save.jac"),
            entrypoint="build",
            args=[],
            session=session,
        )

        cli.enter(
            filename=self.fixture_abs_path("traversing_save.jac"),
            entrypoint="view",
            args=[],
            session=session,
        )

        self.assertEqual(
            "digraph {\n"
            'node [style="filled", shape="ellipse", fillcolor="invis", fontcolor="black"];\n'
            '0 -> 1  [label=""];\n'
            '1 -> 2  [label=""];\n'
            '0 [label="Root()"fillcolor="#FFE9E9"];\n'
            '1 [label="A()"fillcolor="#F0FFF0"];\n'
            '2 [label="B()"fillcolor="#F5E5FF"];\n}',
            self.capturedOutput.getvalue().strip(),
        )

        self._del_session(session)

    def test_custom_access_validation(self) -> None:
        """Test custom access validation."""
        global session
        session = self.fixture_abs_path("custom_access_validation.session")

        ##############################################
        #              CREATE OTHER ROOT             #
        ##############################################

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="create_other_root",
            args=[],
            session=session,
        )

        other_root = self.capturedOutput.getvalue().strip()

        ##############################################
        #                 CREATE NODE                #
        ##############################################

        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="create",
            args=[],
            session=session,
        )
        node = self.capturedOutput.getvalue().strip()

        ##############################################
        #                 CHECK NODE                 #
        ##############################################

        # BY OWNER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            node=node,
        )

        self.assertEqual(
            "A(val1='NO_ACCESS', val2=0)", self.capturedOutput.getvalue().strip()
        )

        # BY OTHER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            root=other_root,
            node=node,
        )

        self.assertEqual("", self.capturedOutput.getvalue().strip())

        ##############################################
        #       UPDATE NODE (GIVE READ ACCESS)       #
        ##############################################

        # UPDATE BY OWNER
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="update",
            args=["READ", None],
            session=session,
            node=node,
        )

        # CHECK BY OTHER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            root=other_root,
            node=node,
        )

        self.assertEqual(
            "A(val1='READ', val2=0)", self.capturedOutput.getvalue().strip()
        )

        ##############################################
        #     UPDATE NODE (BUT STILL READ ACCESS)    #
        ##############################################

        # UPDATE BY OTHER
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="update",
            args=[None, 1],
            session=session,
            root=other_root,
            node=node,
        )

        # CHECK BY OTHER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            root=other_root,
            node=node,
        )

        self.assertEqual(
            "A(val1='READ', val2=0)", self.capturedOutput.getvalue().strip()
        )

        ##############################################
        #       UPDATE NODE (GIVE WRITE ACCESS)      #
        ##############################################

        # UPDATE BY OWNER
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="update",
            args=["WRITE", None],
            session=session,
            node=node,
        )

        # UPDATE BY OTHER
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="update",
            args=[None, 2],
            session=session,
            root=other_root,
            node=node,
        )

        # CHECK BY OTHER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            root=other_root,
            node=node,
        )

        self.assertEqual(
            "A(val1='WRITE', val2=2)", self.capturedOutput.getvalue().strip()
        )

        ##############################################
        #         UPDATE NODE (REMOVE ACCESS)        #
        ##############################################

        # UPDATE BY OWNER
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="update",
            args=["NO_ACCESS", None],
            session=session,
            node=node,
        )

        # UPDATE BY OTHER
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="update",
            args=[None, 5],
            session=session,
            root=other_root,
            node=node,
        )

        # CHECK BY OTHER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            root=other_root,
            node=node,
        )

        self.assertEqual("", self.capturedOutput.getvalue().strip())

        # CHECK BY OWNER
        self._output2buffer()
        cli.enter(
            filename=self.fixture_abs_path("custom_access_validation.jac"),
            entrypoint="check",
            args=[],
            session=session,
            node=node,
        )

        self.assertEqual(
            "A(val1='NO_ACCESS', val2=2)", self.capturedOutput.getvalue().strip()
        )



"""Definition-Use Analysis Pass for the Jac compiler.

This pass performs comprehensive symbol resolution by:
1. Populating symbol tables with additional symbols from various AST constructs:
   - Variable assignments and declarations
   - Function/ability parameters
   - Archetype reference chains
   - Loop variables and comprehension targets
   - With-statement variables

2. Establishing bidirectional links between:
   - Symbol definitions in the symbol table
   - Symbol uses throughout the AST

3. Handling special cases:
   - Inheriting symbol tables from base classes for archetypes
   - Setting appropriate context for symbols in different operations (e.g., delete statements)
   - Marking walker-specific statements (visit, ignore, disengage) when in walker context

This pass is crucial for type checking, access control validation, and code generation as it
creates the complete symbol resolution map for the program.
"""

import jaclang.compiler.unitree as uni
from jaclang.compiler.constant import Tokens as Tok
from jaclang.compiler.passes import UniPass


class DefUsePass(UniPass):
    """Jac Ast build pass."""

    def enter_archetype(self, node: uni.Archetype) -> None:
        node.sym_tab.inherit_baseclasses_sym(node)

        def inform_from_walker(node: uni.UniNode) -> None:
            for i in (
                node.get_all_sub_nodes(uni.VisitStmt)
                + node.get_all_sub_nodes(uni.DisengageStmt)
                + node.get_all_sub_nodes(uni.EdgeOpRef)
                + node.get_all_sub_nodes(uni.EventSignature)
            ):
                i.from_walker = True

        if node.arch_type.name == Tok.KW_WALKER:
            inform_from_walker(node)
            for i in self.get_all_sub_nodes(node, uni.Ability):
                if isinstance(i.body, uni.ImplDef):
                    inform_from_walker(i.body)

    def enter_enum(self, node: uni.Enum) -> None:
        node.sym_tab.inherit_baseclasses_sym(node)

    def enter_type_ref(self, node: uni.TypeRef) -> None:
        node.sym_tab.use_lookup(node)

    def enter_param_var(self, node: uni.ParamVar) -> None:
        node.sym_tab.def_insert(node)

    def enter_has_var(self, node: uni.HasVar) -> None:
        if isinstance(node.parent, uni.ArchHas):
            node.sym_tab.def_insert(
                node, single_decl="has var", access_spec=node.parent
            )
        else:
            self.ice("Inconsistency in AST, has var should be under arch has")

    def enter_assignment(self, node: uni.Assignment) -> None:
        for i in node.target:
            if isinstance(i, uni.AtomTrailer):
                i.sym_tab.chain_def_insert(i.as_attr_list)
            elif isinstance(i, uni.AstSymbolNode):
                i.sym_tab.def_insert(i)
            else:
                self.log_error("Assignment target not valid")

    def enter_inner_compr(self, node: uni.InnerCompr) -> None:
        if isinstance(node.target, uni.AtomTrailer):
            node.target.sym_tab.chain_def_insert(node.target.as_attr_list)

        elif isinstance(node.target, uni.AstSymbolNode):
            node.target.sym_tab.def_insert(node.target)
        else:
            self.log_error("Named target not valid")

    def enter_atom_trailer(self, node: uni.AtomTrailer) -> None:
        chain = node.as_attr_list
        node.sym_tab.chain_use_lookup(chain)

    def enter_special_var_ref(self, node: uni.SpecialVarRef) -> None:
        node.sym_tab.use_lookup(node)

    def enter_float(self, node: uni.Float) -> None:
        node.sym_tab.use_lookup(node)

    def enter_int(self, node: uni.Int) -> None:
        node.sym_tab.use_lookup(node)

    def enter_string(self, node: uni.String) -> None:
        node.sym_tab.use_lookup(node)

    def enter_bool(self, node: uni.Bool) -> None:
        node.sym_tab.use_lookup(node)

    def enter_builtin_type(self, node: uni.BuiltinType) -> None:
        node.sym_tab.use_lookup(node)

    def enter_name(self, node: uni.Name) -> None:
        if not isinstance(node.parent, uni.AtomTrailer):
            node.sym_tab.use_lookup(node)

    def enter_in_for_stmt(self, node: uni.InForStmt) -> None:
        if isinstance(node.target, uni.AtomTrailer):
            node.target.sym_tab.chain_def_insert(node.target.as_attr_list)
        elif isinstance(node.target, uni.AstSymbolNode):
            node.target.sym_tab.def_insert(node.target)
        else:
            self.log_error("For loop assignment target not valid")

    def enter_expr_as_item(self, node: uni.ExprAsItem) -> None:
        if node.alias:
            if isinstance(node.alias, uni.AtomTrailer):
                node.alias.sym_tab.chain_def_insert(node.alias.as_attr_list)
            elif isinstance(node.alias, uni.AstSymbolNode):
                node.alias.sym_tab.def_insert(node.alias)
            else:
                self.log_error("For expr as target not valid")



"""Test pass module."""

import jaclang.compiler.unitree as uni
from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCase


class SemDefMatchPassTests(TestCase):
    """Test pass module."""

    def test_sem_def_match(self) -> None:
        """Basic test for pass."""
        (out := JacProgram()).compile(self.fixture_abs_path("sem_def_match.jac"))
        self.assertFalse(out.errors_had)
        mod = out.mod.hub[self.fixture_abs_path("sem_def_match.jac")]

        self.assertEqual(mod.lookup("E").decl.name_of.semstr, "An enum representing some values.")  # type: ignore
        self.assertEqual(mod.lookup("E").fetch_sym_tab.lookup("A").decl.name_of.semstr, "The first value of the enum E.")  # type: ignore
        self.assertEqual(mod.lookup("E").fetch_sym_tab.lookup("B").decl.name_of.semstr, "The second value of the enum E.")  # type: ignore

        self.assertEqual(mod.lookup("Person").decl.name_of.semstr, "A class representing a person.")  # type: ignore

        person_scope = mod.lookup("Person").fetch_sym_tab  # type: ignore
        self.assertEqual(person_scope.lookup("name").decl.name_of.semstr, "The name of the person.")  # type: ignore
        self.assertEqual(person_scope.lookup("yob").decl.name_of.semstr, "The year of birth of the person.")  # type: ignore

        sym_calc_age = person_scope.lookup("calc_age")  # type: ignore
        self.assertEqual(sym_calc_age.decl.name_of.semstr, "Calculate the age of the person.")  # type: ignore

        calc_age_scope = sym_calc_age.fetch_sym_tab  # type: ignore
        self.assertEqual(calc_age_scope.lookup("year").decl.name_of.semstr, "The year to calculate the age against.")  # type: ignore

        self.assertEqual(mod.lookup("OuterClass").decl.name_of.semstr, "A class containing an inner class.")  # type: ignore
        outer_scope = mod.lookup("OuterClass").fetch_sym_tab  # type: ignore
        self.assertEqual(outer_scope.lookup("InnerClass").decl.name_of.semstr, "An inner class within OuterClass.")  # type: ignore
        inner_scope = outer_scope.lookup("InnerClass").fetch_sym_tab  # type: ignore
        self.assertEqual(inner_scope.lookup("inner_value").decl.name_of.semstr, "A value specific to the inner class.")  # type: ignore




"""Jac compiler tools."""

import contextlib
import logging
import os
import shutil
import sys

from jaclang.vendor.lark.tools import standalone


def generate_static_parser(force: bool = False) -> None:
    """Generate static parser."""
    cur_dir = os.path.dirname(__file__)
    if force or not os.path.exists(os.path.join(cur_dir, "larkparse", "jac_parser.py")):
        if os.path.exists(os.path.join(cur_dir, "larkparse")):
            shutil.rmtree(os.path.join(cur_dir, "larkparse"))
        os.makedirs(os.path.join(cur_dir, "larkparse"), exist_ok=True)
        with open(os.path.join(cur_dir, "larkparse", "__init__.py"), "w"):
            pass
        save_argv = sys.argv
        sys.argv = [
            "lark",
            os.path.join(cur_dir, "jac.lark"),
            "-o",
            os.path.join(cur_dir, "larkparse", "jac_parser.py"),
            "-c",
        ]
        standalone.main()
        sys.argv = save_argv


try:
    from jaclang.compiler.larkparse import jac_parser as jac_lark
except ModuleNotFoundError:
    generate_static_parser(force=True)
    from jaclang.compiler.larkparse import jac_parser as jac_lark

jac_lark.logger.setLevel(logging.DEBUG)
contextlib.suppress(ModuleNotFoundError)

TOKEN_MAP = {
    x.name: x.pattern.value
    for x in jac_lark.Lark_StandAlone().parser.lexer_conf.terminals
}

# fmt: off
TOKEN_MAP.update(
    {
        "CARROW_L": "<++", "CARROW_R": "++>", "GLOBAL_OP": "global",
        "NONLOCAL_OP": "nonlocal", "WALKER_OP": ":walker:", "NODE_OP": ":node:",
        "EDGE_OP": ":edge:", "CLASS_OP": ":class:", "OBJECT_OP": ":obj:",
        "TYPE_OP": "`", "ABILITY_OP": ":can:", "NULL_OK": "?",
        "KW_OR": "|", "ARROW_BI": "<-->", "ARROW_L": "<--",
        "ARROW_R": "-->", "ARROW_L_P1": "<-:", "ARROW_R_P2": ":->",
        "ARROW_L_P2": ":<-", "ARROW_R_P1": "->:", "CARROW_BI": "<++>",
        "CARROW_L_P1": "<+:", "RSHIFT_EQ": ">>=", "ELLIPSIS": "...",
        "CARROW_R_P2": ":+>", "CARROW_L_P2": ":<+", "CARROW_R_P1": "+>:",
        "PIPE_FWD": "|>", "PIPE_BKWD": "<|", "A_PIPE_FWD": ":>",
        "A_PIPE_BKWD": "<:", "DOT_FWD": ".>", "STAR_POW": "**",
        "STAR_MUL": "*", "FLOOR_DIV": "//", "DIV": "/",
        "PYNLINE": "::py::", "ADD_EQ": "+=", "SUB_EQ": "-=",
        "STAR_POW_EQ": "**=", "MUL_EQ": "*=", "FLOOR_DIV_EQ": "//=",
        "DIV_EQ": "/=", "MOD_EQ": "%=", "BW_AND_EQ": "&=",
        "BW_OR_EQ": "|=", "BW_XOR_EQ": "^=", "BW_NOT_EQ": "~=",
        "LSHIFT_EQ": "<<=",
    }
)
# fmt: on


__all__ = ["jac_lark", "TOKEN_MAP"]



"""Test ast build pass module."""

import ast as ast3
from difflib import unified_diff


from jaclang.compiler.passes.tool import JacFormatPass
from jaclang.compiler.program import JacProgram
from jaclang.utils.test import TestCaseMicroSuite


class JacUnparseTests(TestCaseMicroSuite):
    """Test pass module."""

    TargetPass = JacFormatPass

    def test_double_unparse(self) -> None:
        """Parse micro jac file."""
        try:
            code_gen_pure = JacProgram().compile(
                self.examples_abs_path("manual_code/circle.jac")
            )
            x = code_gen_pure.unparse()
            y = code_gen_pure.unparse()
            self.assertEqual(x, y)
        except Exception as e:
            print("\n".join(unified_diff(x.splitlines(), y.splitlines())))
            raise e

    def micro_suite_test(self, filename: str) -> None:
        """Parse micro jac file."""
        code_gen_pure = JacProgram().compile(
            self.fixture_abs_path(filename),
        )
        before = ast3.dump(code_gen_pure.gen.py_ast[0], indent=2)
        x = code_gen_pure.unparse()
        code_gen_jac = JacProgram().compile(
            use_str=x,
            file_path=filename,
        )
        after = ast3.dump(code_gen_jac.gen.py_ast[0], indent=2)
        if "circle_clean_tests.jac" in filename:
            self.assertEqual(
                len(
                    [
                        i
                        for i in unified_diff(
                            before.splitlines(), after.splitlines(), n=0
                        )
                        if "test" not in i
                    ]
                ),
                5,
            )
        else:
            try:
                self.assertEqual(
                    len(
                        "\n".join(unified_diff(before.splitlines(), after.splitlines()))
                    ),
                    0,
                )
            except Exception as e:
                print(
                    "\n".join(
                        unified_diff(before.splitlines(), after.splitlines(), n=10)
                    )
                )
                raise e


JacUnparseTests.self_attach_micro_tests()
